#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-11-21 22:40:17
Last Modified by: Hanyu Wang
Last Modified time: 2023-11-21 23:43:54
"""

from xyz.circuit import QCircuit
from xyz.qstate import QState
from .library import DefaultLibrary, Library


def library_cnot_synthesis(
    circuit: QCircuit,
    state: QState,
    library: Library = DefaultLibrary(),
):
    """Synthesize the quantum circuit .

    :param circuit: [description]
    :type circuit: QCircuit
    :param state: [description]
    :type state: QState
    :param library: [description], defaults to DefaultLibrary()
    :type library: Library, optional
    :return: [description]
    :rtype: [type]
    """
    library.initialize_queue(state, circuit)
    library.explore()
    gates = library.get_solution()
    return gates



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-06-22 14:45:56
Last Modified by: Hanyu Wang
Last Modified time: 2023-06-23 00:09:04
"""

from .control import *
from .rotation import *
from .qubit import *
from .gate import *
from .unitary import *



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-09-01 12:51:03
Last Modified by: Hanyu Wang
Last Modified time: 2023-09-01 12:56:17
"""

import numpy as np
import copy

from queue import PriorityQueue

import xyz.qstate as qs
import xyz.operator as op
import xyz.circuit as qc

from ._ground_state_calibration import ground_state_calibration

DEPRECATED: bool = True

def exact_cnot_synthesis_legacy(
    circuit: qc.QCircuit,
    target_state: qs.QState,
    verbose_level: int = 0,
    cnot_limit: int = None,
):
    """This function finds the exact cnot_cnot_synthesis of a circuit .

    :param circuit: [description]
    :type circuit: QCircuit
    :param qubit_mapping: [description]
    :type qubit_mapping: dict
    :param target_state: [description]
    :type target_state: qs.QState
    :param optimality_level: [description], defaults to 3
    :type optimality_level: int, optional
    :param verbose_level: [description], defaults to 0
    :type verbose_level: int, optional
    :raises ValueError: [description]
    :return: [description]
    :rtype: [type]
    """

    # we should raise an error, since this is the old version
    if DEPRECATED:
        raise ValueError(
            "This is the old version of exact_cnot_synthesis, please use the new version"
        )

    # now we start the search
    visited_states = set()
    state_queue = PriorityQueue()
    enquened_states = {}
    record = {}

    def explore_state(
        curr_state: qs.QState, gates: op.QOperator, curr_cost: int
    ) -> qs.QState:
        """Explore a state in a SRGraph ."""
        nonlocal visited_states, state_queue, enquened_states, record, circuit

        next_state = copy.deepcopy(curr_state)
        gate_cost = 0
        for gate in gates:
            try:
                match gate.qgate_type:
                    case qc.QGateType.X:
                        gate: qc.X
                        next_state = next_state.apply_x(gate.target_qubit.index)
                        gate_cost += 0.05
                    case qc.QGateType.CX:
                        gate: qc.CX
                        next_state = next_state.apply_cx(
                            gate.control_qubit.index,
                            gate.phase,
                            gate.target_qubit.index,
                        )
                        gate_cost += 1
                    case qc.QGateType.RY:
                        gate: qc.RY
                        next_state = next_state.apply_ry(
                            gate.target_qubit.index, -gate.theta
                        )
                        gate_cost += 0.05
                    case _:
                        raise ValueError(f"Unknown gate type: {gate.qgate_type}")

                if verbose_level >= 3:
                    print(
                        f"next_state: {next_state}, curr_state = {curr_state} gate = {gate}"
                    )
            except ValueError:
                return None

        next_cost = curr_cost + gate_cost
        next_state_repr = next_state.repr()

        # we skip the state if it is already visited
        if next_state_repr in visited_states:
            return None

        # we skip the state if it is already enquened and the cost is higher
        if (
            next_state_repr in enquened_states
            and next_cost >= enquened_states[next_state_repr]
        ):
            return None

        # now we add the state to the queue
        state_queue.put((next_cost, next_state))
        enquened_states[next_state_repr] = next_cost

        # and record the quantum_operator
        if verbose_level >= 3:
            print(
                f"recording [{hash(next_state)}] <- {hash(curr_state)}, gate: {gates}"
            )
        record[hash(next_state)] = hash(curr_state), gates[::-1]

        return next_state

    # begin of the exact synthesis algorithm
    num_qubits = target_state.num_qubits
    initial_state = qs.QState.ground_state(num_qubits)

    curr_state = target_state
    curr_cost = 0
    state_queue.put((curr_cost, curr_state))

    solution_reached: bool = False

    while not state_queue.empty():
        curr_cost, curr_state = state_queue.get()

        if verbose_level == 1:
            print(
                f"#visited: {len(visited_states)} #enquened: {state_queue.qsize()}, #cost: {curr_cost}\r",
                end="",
            )

        if verbose_level >= 2:
            print(f"\n\ncurr_state: {curr_state}, cost: {curr_cost}")

        if cnot_limit is not None and curr_cost.cnot_cost > cnot_limit:
            # this will then raise an ValueError
            break

        if curr_state == initial_state:
            # then we have found the solution
            solution_reached = True
            break

        _sparsity = curr_state.get_sparsity()

        if _sparsity == 1:
            solution_reached = True
            break

        curr_state_repr = curr_state.repr()
        visited_states.add(curr_state_repr)
        supports = curr_state.get_supports()

        for target_qubit in supports:
            # X
            explore_state(curr_state, [qc.X(circuit.qubit_at(target_qubit))], curr_cost)

            thetas = curr_state.get_ry_angles(target_qubit)
            if verbose_level >= 3:
                print(f"thetas: {thetas}")

            # RY
            # find the most frequent theta
            # best_theta = curr_state.get_most_frequent_theta(target_qubit)

            # if best_theta is not None and not np.isclose(best_theta, 0):
            for theta in sorted(thetas):
                explore_state(
                    curr_state,
                    [qc.RY(theta, circuit.qubit_at(target_qubit))],
                    curr_cost,
                )
                continue

            # CX
            for control_qubit in supports:
                if control_qubit != target_qubit:
                    cry_thetas = curr_state.get_cry_angles(control_qubit, target_qubit)

                    prev_theta = None
                    for theta in sorted(cry_thetas):
                        if prev_theta is not None and np.isclose(theta, prev_theta):
                            continue
                        if verbose_level >= 3:
                            print(f"cry theta: {theta}")
                        explore_state(
                            curr_state,
                            [
                                qc.RY(theta, circuit.qubit_at(target_qubit)),
                                qc.CX(
                                    circuit.qubit_at(control_qubit),
                                    0,
                                    circuit.qubit_at(target_qubit),
                                ),
                                qc.RY(-theta, circuit.qubit_at(target_qubit)),
                            ],
                            curr_cost,
                        )
                        explore_state(
                            curr_state,
                            [
                                qc.RY(theta, circuit.qubit_at(target_qubit)),
                                qc.CX(
                                    circuit.qubit_at(control_qubit),
                                    1,
                                    circuit.qubit_at(target_qubit),
                                ),
                                qc.RY(-theta, circuit.qubit_at(target_qubit)),
                            ],
                            curr_cost,
                        )
    if not solution_reached:
        raise ValueError("No solution found")

    final_state = qs.QState(curr_state.index_to_weight, curr_state.num_qubits)

    x_gates = ground_state_calibration(circuit, final_state)

    if verbose_level >= 2:
        print("\n\n")
    for record_key, record_value in record.items():
        prev_state, gate = record_value
        if verbose_level >= 2:
            print(
                f"record_key: {record_key}\n\t prev_state: {prev_state}\n\t gate: {gate}"
            )

    gates = x_gates[:]
    backtraced_states: set = set()
    curr_hash = hash(curr_state)
    while curr_hash in record:
        if curr_hash in backtraced_states:
            raise ValueError("Loop found")
        backtraced_states.add(curr_hash)
        prev_hash, _gates = record[curr_hash]
        for _gate in _gates:
            gates.append(_gate)
        curr_hash = prev_hash
        if verbose_level >= 1:
            print(f"curr_hash: {curr_hash}, gate: {gate}")

    return gates



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-09-15 13:31:48
Last Modified by: Hanyu Wang
Last Modified time: 2023-09-15 13:37:33
"""


def index_to_str(index_to_weight: dict, num_qubits: int):
    """Prints the index_to_weight dictionary.

    :param index_to_weight: [description]
    :type index_to_weight: dict
    """

    index_str = ""

    if len(index_to_weight) == 0:
        return index_str

    if index_to_weight is None:
        return index_str

    assert isinstance(
        index_to_weight, dict
    ), f"index_to_weight must be a dictionary, got {type(index_to_weight)}"

    for index, weight in index_to_weight.items():
        index_str += f"{index:0{num_qubits}b}: {weight:0.02f}, "

    return index_str



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-06-28 14:34:30
Last Modified by: Hanyu Wang
Last Modified time: 2023-06-28 14:40:59
"""

from .algorithms import *
from .operator import *
from .qstate import *
from .utils import *
from .boolean import *
from .circuit import *



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-06-17 07:59:34
Last Modified by: Hanyu Wang
Last Modified time: 2023-06-17 07:59:44
"""

import numpy as np


def W_state(num_qubits: int) -> np.ndarray:
    """Generate the W state for the W gate .

    :param num_qubits: [description]
    :type num_qubits: int
    :return: [description]
    :rtype: np.ndarray
    """
    state = np.zeros(2**num_qubits)

    for i in range(num_qubits):
        state[(2**i)] = 1

    return state / np.sqrt(num_qubits)



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-06-28 10:38:39
Last Modified by: Hanyu Wang
Last Modified time: 2023-06-28 10:55:26
"""

from .qstate import *
from .common import *
from .index_to_weight import *



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-06-22 14:39:56
Last Modified by: Hanyu Wang
Last Modified time: 2023-06-22 23:25:35
"""

from typing import List

import numpy as np


class RotationGate:
    """Classmethod to handle RotationGate ."""

    def __init__(self, theta: float) -> None:
        self.theta = theta

    def is_trivial(self) -> bool:
        """Whether the model is trivial to be trivial .

        :return: [description]
        :rtype: bool
        """
        return np.isclose(self.theta, 0) or np.isclose(self.theta, 2 * np.pi)

    def is_pi(self) -> bool:
        """True if theta is a pi - > pi .

        :return: [description]
        :rtype: bool
        """
        return np.isclose(self.theta, np.pi) or np.isclose(self.theta, -np.pi)

    def get_theta(self) -> float:
        """Get the theta of the rotation gate .

        :return: [description]
        :rtype: float
        """
        return self.theta


class MultiRotationGate:
    """Class method for creating a new rotation gate ."""

    def __init__(self, thetas: List[float]) -> None:
        self.thetas = thetas



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-08-31 14:04:53
Last Modified by: Hanyu Wang
Last Modified time: 2023-08-31 14:24:19
"""

from .truth_table import *
from .sum_of_product import *



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-09-13 10:29:33
Last Modified by: Hanyu Wang
Last Modified time: 2023-09-13 10:43:53
"""

import numpy as np

from .base import BasicGate, UnitaryGate, ControlledGate, QBit, QGateType


class CU(BasicGate, UnitaryGate, ControlledGate):
    """Classmethod to create a CU and ControlledGate .

    :param BasicGate: [description]
    :type BasicGate: [type]
    :param ControlledGate: [description]
    :type ControlledGate: [type]
    """

    def __init__(
        self, unitary: np.ndarray, control_qubit: QBit, phase: int, target_qubit: QBit
    ) -> None:
        BasicGate.__init__(self, QGateType.CU, target_qubit)
        UnitaryGate.__init__(self, unitary)
        ControlledGate.__init__(self, control_qubit, phase)

    def __str__(self) -> str:
        phase_str = "" if self.phase == 1 else "~"
        return f"C({phase_str}{self.control_qubit})U({self.target_qubit})"

    def get_cnot_cost(self) -> int:
        """Returns the cost of the cost of the gate.

        :return: [description]
        :rtype: int
        """
        return 2



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-06-22 22:43:37
Last Modified by: Hanyu Wang
Last Modified time: 2023-06-23 00:14:40
"""

from collections import namedtuple

from typing import List
import numpy as np

from xyz.circuit.basic_gates.ry import RY
from xyz.utils import call_with_global_timer

from .basic_gates import QGate, QGateType, MULTIPLEXY, CRX, CU, MCRY, X, MCMY
from .decomposition import decompose_mcry, control_sequence_to_gates


def __map_muxy(gate: MULTIPLEXY) -> List[QGate]:
    """Convert a multi - gate into a list of gates .

    :param gate: [description]
    :type gate: MULTIPLEXY
    :return: [description]
    :rtype: List[QGate]
    """
    assert gate.type == QGateType.MULTIPLEX_Y
    rotation_table = [gate.theta0, gate.theta1]

    control_sequence = decompose_mcry(rotation_table)

    gates = control_sequence_to_gates(
        control_sequence, [gate.control_qubit], gate.target_qubit
    )

    return gates


def __map_mcry(gate: QGate) -> List[QGate]:
    """Convert a MCRY gate into a list of gates .

    :param gate: [description]
    :type gate: QGate
    :raises Exception: [description]
    :return: [description]
    :rtype: List[QGate]
    """
    match gate.get_qgate_type():
        case QGateType.MCRY:
            control_qubits = gate.control_qubits
            phases = gate.phases
            target_qubit = gate.target_qubit

        case QGateType.CRY:
            control_qubits = [gate.control_qubit]
            phases = [gate.phase]
            target_qubit = gate.target_qubit

        case _:
            raise ValueError("Not a MCRY gate")

    # we prepare the rotation table
    rotation_table = np.zeros(2 ** (len(control_qubits)))

    rotated_index = 0
    for i, controlled_by_one in enumerate(phases):
        if controlled_by_one is True:
            rotated_index += 2**i

    # only rotate the target qubit if the control qubits are in the positive phase
    rotation_table[rotated_index] = gate.theta

    control_sequence = decompose_mcry(rotation_table)

    gates = control_sequence_to_gates(control_sequence, control_qubits, target_qubit)

    return gates


def __map_mcmy(gate: MCMY) -> List[QGate]:
    """Convert a MCMY gate into a list of gates .

    :param gate: [description]
    :type gate: MCMY
    :return: [description]
    :rtype: List[QGate]
    """

    rotation_table = gate.rotation_table

    control_sequence = decompose_mcry(rotation_table=rotation_table)

    gates = control_sequence_to_gates(
        control_sequence,
        gate.control_qubits,
        gate.target_qubit,
    )

    return gates


def theta_to_unitary(theta: float):
    """Converts theta to a unitary unitary gate .

    :param theta: [description]
    :type theta: float
    """
    raw_matrix = np.array(
        [
            [np.cos(theta / 2), -np.sin(theta / 2)],
            [np.sin(theta / 2), np.cos(theta / 2)],
        ]
    )
    return raw_matrix


def unitary_convert(unitary: np.ndarray, coef: float, signal: int):
    """Convert a vector in a unitary unitary function to a unitary function .

    :param unitary: [description]
    :type unitary: np.ndarray
    :param coef: [description]
    :type coef: float
    :param signal: [description]
    :type signal: int
    :return: [description]
    :rtype: [type]
    """
    param = 1 / np.abs(coef)

    values, vectors = np.linalg.eig(unitary)
    updated_unitary = (
        np.power(values[0] + 0j, param) * vectors[:, [0]] @ vectors[:, [0]].conj().T
    )
    updated_unitary = (
        updated_unitary
        + np.power(values[1] + 0j, param) * vectors[:, [1]] @ vectors[:, [1]].conj().T
    )

    if signal < 0:
        updated_unitary = np.linalg.inv(updated_unitary)

    return updated_unitary


def __map_mcry_linear(mcry_gate: MCRY) -> List[QGate]:
    """Map a gate to the circuit .

    :param gate: [description]
    :type gate: QGate
    :return: [description]
    :rtype: List[QGate]
    """

    assert mcry_gate.get_qgate_type() == QGateType.MCRY

    # we first preprocess the relavant qubits
    qubit_list = mcry_gate.control_qubits + [mcry_gate.target_qubit]
    num_qubits = len(qubit_list)

    # special case for single qubit
    if num_qubits == 1:
        gates = []
        gate = RY(mcry_gate.theta, qubit_list[0])
        gates.append(gate)
        return gates

    # special case for two qubits
    # if num_qubits == 2:
    #     gates = __map_mcry(mcry_gate)
    #     return gates

    # now we handle the general case
    def convert_rec(
        unitary: np.ndarray, num_qubits: int, is_first: bool = True, step: int = 1
    ):
        """Convert a unitary unitary to a unitary gate .

        :param unitary: [description]
        :type unitary: np.ndarray
        :param num_qubits: [description]
        :type num_qubits: int
        :param is_first: [description], defaults to True
        :type is_first: bool, optional
        :param step: [description], defaults to 1
        :type step: int, optional
        :return: [description]
        :rtype: [type]
        """

        nonlocal qubit_list

        pairs = namedtuple("pairs", ["control", "target"])

        if step == 1:
            start = 0
            reverse = True

        else:
            start = 1
            reverse = False

        qubit_pairs = [
            pairs(control, target)
            for target in range(num_qubits)
            for control in range(start, target)
        ]

        qubit_pairs.sort(key=lambda e: e.control + e.target, reverse=reverse)

        gates = []
        for pair in qubit_pairs:
            exponent = pair.target - pair.control
            if pair.control == 0:
                exponent = exponent - 1
            param = 2**exponent
            signal = -1 if (pair.control == 0 and not is_first) else 1
            signal = signal * step

            if pair.target == num_qubits - 1 and is_first:
                updated_unitary = unitary_convert(unitary, param, signal)
                gate = CU(
                    updated_unitary,
                    qubit_list[pair.control],
                    True,
                    qubit_list[pair.target],
                )
                gates.append(gate)
            else:
                gate = CRX(
                    signal * np.pi / param,
                    qubit_list[pair.control],
                    True,
                    qubit_list[pair.target],
                )
                gates.append(gate)

        return gates

    unitary = theta_to_unitary(mcry_gate.get_theta())

    gates = []
    # frist we consider the control phases
    for phase, control_qubit in zip(mcry_gate.get_phases(), mcry_gate.control_qubits):
        if phase == 0:
            gate = X(control_qubit)
            gates.append(gate)

    gates_c1 = convert_rec(unitary, num_qubits)
    gates_c2 = convert_rec(unitary, num_qubits, step=-1)
    gates_c3 = convert_rec(unitary, num_qubits - 1, is_first=False)
    gates_c4 = convert_rec(unitary, num_qubits - 1, is_first=False, step=-1)

    for gate in gates_c1:
        gates.append(gate)
    for gate in gates_c2:
        gates.append(gate)
    for gate in gates_c3:
        gates.append(gate)
    for gate in gates_c4:
        gates.append(gate)

    # frist we consider the control phases
    for phase, control_qubit in zip(mcry_gate.get_phases(), mcry_gate.control_qubits):
        if phase == 0:
            gate = X(control_qubit)
            gates.append(gate)

    return gates


@call_with_global_timer
def add_gate_mapped(self, gate: QGate) -> None:
    """Add a gate to the circuit .

    :param gate: [description]
    :type gate: QGate
    """
    if not self.map_gates:
        self.append_gate(gate)
        return

    match gate.qgate_type:
        case QGateType.MULTIPLEX_Y:
            gates = __map_muxy(gate)
            self.append_gates(gates)

        case QGateType.MCRY:
            if len(gate.get_control_qubits()) >= 8:
                gates = __map_mcry_linear(gate)
            else:
                gates = __map_mcry(gate)

            # this may cause problem if the gates returned are still MCRYs
            self.add_gates(gates)

        case QGateType.MCMY:
            gates = __map_mcmy(gate)

            self.add_gates(gates)

        case QGateType.CRY:
            gates = __map_mcry(gate)
            self.append_gates(gates)

        case _:
            self.append_gate(gate)



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-06-22 14:33:21
Last Modified by: Hanyu Wang
Last Modified time: 2023-06-22 23:43:32
"""

from typing import List

from .basic_gates import QGate, QGateType, RY, CRY
from ._mapping import add_gate_mapped


def _add_gate_optimized(self, gate: QGate) -> None:
    """Add a new gate to this gate .

    :param gate: [description]
    :type gate: QGate
    """

    match gate.get_qgate_type():
        case QGateType.CU:
            if gate.is_z_trivial():
                reduced_gate = CRY(
                    gate.get_beta(),
                    gate.control_qubit,
                    gate.get_phase(),
                    gate.target_qubit,
                )
                _add_gate_optimized(self, reduced_gate)
            elif gate.is_z_trivial_not():
                reduced_gate = CRY(
                    -gate.get_beta(),
                    gate.control_qubit,
                    gate.get_phase(),
                    gate.target_qubit,
                )
                _add_gate_optimized(self, reduced_gate)
            else:
                add_gate_mapped(self, gate)

        case QGateType.RY:
            if gate.is_trivial():
                return
            else:
                add_gate_mapped(self, gate)

        case QGateType.CRY:
            add_gate_mapped(self, gate)
        case QGateType.MCRY:
            if gate.has_zero_controls():
                reduced_gate = RY(gate.theta, gate.target_qubit)
                _add_gate_optimized(self, reduced_gate)

            # other wise we can use the same method as CRY
            elif gate.has_one_control():
                reduced_gate = CRY(
                    gate.theta,
                    gate.control_qubits[0],
                    gate.phases[0],
                    gate.target_qubit,
                )
                _add_gate_optimized(self, reduced_gate)

            else:
                add_gate_mapped(self, gate)
        case _:
            add_gate_mapped(self, gate)


def _add_gates_optimized(self, gates: List[QGate]) -> None:
    """
    Add a list of gates to the circuit, with optimization
    """
    for gate in gates:
        _add_gate_optimized(self, gate)



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-06-25 12:02:38
Last Modified by: Hanyu Wang
Last Modified time: 2023-06-25 12:04:43
"""

from .w_state import *
from .dicke_state import *
from .ghz_state import *
from .c_state import *



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-09-15 13:25:54
Last Modified by: Hanyu Wang
Last Modified time: 2023-09-15 13:28:55
"""

from .diff_states import get_difference
from .verify_circuit import verify_circuit_and_count_cnot, simulate_circuit



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-06-17 08:11:03
Last Modified by: Hanyu Wang
Last Modified time: 2023-06-17 08:11:27
"""

from .colors import *
from .timer import *



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-08-31 12:44:45
Last Modified by: Hanyu Wang
Last Modified time: 2023-08-31 12:46:16
"""

from typing import List
import numpy as np

from scipy.linalg import solve

from xyz.utils import call_with_global_timer
from ..basic_gates import QGate, RY, CX, QBit


@call_with_global_timer
def find_thetas(alphas):
    """Find theta matrix for the given alphas .

    :param alphas: [description]
    :type alphas: [type]
    :return: [description]
    :rtype: [type]
    """
    size = len(alphas)
    # for the gray code matrix
    gray_code_coefficient_matrix = np.zeros((size, size))
    for i in range(size):
        for j in range(size):
            #  The exponent is the bit-wise inner product of the binary vectors for the standard binary code representation of the integer i (bi) and the binary representation of the i th value of the gray code up to a value of 2) The j th value of the gray code is calculated using the bit-wise XOR of the unsigned binary j and a single shift right of the value of j , like so: ”j XOR (j>>1)” for C++ code

            bitwise_inner_product = bin(i & (j ^ (j >> 1))).count("1")
            gray_code_coefficient_matrix[i, j] = (-1) ** bitwise_inner_product

    thetas = solve(gray_code_coefficient_matrix, alphas)

    return thetas


@call_with_global_timer
def decompose_mcry(rotation_table: list):
    """Synthesize multiple controlled rotations .

    :param rotation_table: [description]
    :type rotation_table: list
    :return: [description]
    :rtype: [type]
    """
    num_controls = int(np.log2(len(rotation_table)))

    assert num_controls > 0, "The number of controls must be greater than 0"

    alphas = rotation_table[:]
    thetas = find_thetas(alphas)

    # return a list of control sequences
    control_sequence: list = []

    prev_gray_code = 0

    # print(f"num_controls: {num_controls}, thetas: {thetas}")

    for i, theta in enumerate(thetas):
        # get the bit that changed in the i of gray code
        # for example, if num_qubits = 3, then the gray code is 000, 001, 011, 010, 110, 111, 101, 100
        # the control id is 0, 1, 0, 2, 0, 1, 0, 2, respectively
        # which is determined by the number of 1 in the binary representation of the gray code
        curr_gray_code = (
            ((i + 1) ^ ((i + 1) >> 1)) if i < (2**num_controls) - 1 else 0
        )

        diff = curr_gray_code ^ prev_gray_code
        # print(f"i: {i}, curr_gray_code: {curr_gray_code}, prev_gray_code: {prev_gray_code}, diff: {diff}")
        prev_gray_code = curr_gray_code

        control_id = int(np.log2(diff))

        control_sequence.append((theta, control_id))

    return control_sequence


@call_with_global_timer
def control_sequence_to_gates(
    control_sequence: list, control_qubits: List[QBit], target_qubit: QBit
) -> List[QGate]:
    """Convert a control sequence into a list of gates .

    :param control_sequence: [description]
    :type control_sequence: list
    :param control_qubits: [description]
    :type control_qubits: List[QBit]
    :param target_qubit: [description]
    :type target_qubit: QBit
    :return: [description]
    :rtype: List[QGate]
    """
    gates: List[QGate] = []

    for control in control_sequence:
        rotation_theta, control_id = control

        gates.append(RY(rotation_theta, target_qubit))
        gates.append(CX(control_qubits[control_id], 1, target_qubit))

    return gates



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-06-21 14:02:28
Last Modified by: Hanyu Wang
Last Modified time: 2023-06-25 12:59:39
"""

from .synthesize import *
from .hybrid_synthesis import *
from ._exact_cnot_synthesis_legacy import exact_cnot_synthesis_legacy
from ._heuristic_cnot_synthesis import heuristic_cnot_synthesis
from .library_cnot_synthesis import library_cnot_synthesis



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-06-22 13:11:53
Last Modified by: Hanyu Wang
Last Modified time: 2023-06-22 15:03:19
"""

from .base import BasicGate, QBit, QGateType, RotationGate


class RY(RotationGate, BasicGate):
    """Constructs a RYYY gate .

    :param RotationGate: [description]
    :type RotationGate: [type]
    :param BasicGate: [description]
    :type BasicGate: [type]
    """

    def __init__(self, theta: float, target_qubit: QBit) -> None:
        BasicGate.__init__(self, QGateType.RY, target_qubit)
        RotationGate.__init__(self, theta)

    def __str__(self) -> str:
        return f"RY({self.theta:0.02f})"

    def get_cnot_cost(self) -> int:
        """Returns the cost of the cost of the gate.

        :return: [description]
        :rtype: int
        """
        return 0



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-06-28 11:34:42
Last Modified by: Hanyu Wang
Last Modified time: 2023-06-28 11:35:15
"""

from enum import Enum, auto


class QOperatorType(Enum):
    """A method to set the operator type for the QOperatorType .

    :param Enum: [description]
    :type Enum: [type]
    """

    MCRY = auto()
    X = auto()
    CX = auto()
    T0 = auto()
    T1 = auto()
    CT0 = auto()
    CT1 = auto()


class QOperator:
    """Class method to create a QOperator class ."""

    def __init__(self, operator_type: QOperatorType) -> None:
        self.operator_type = operator_type


class QOperatorBase:
    """Class method to create a QOperatorBase subclass ."""

    def __init__(self, operator_type: QOperatorType, target_qubit_index: int) -> None:
        self.operator_type = operator_type
        self.target_qubit_index = target_qubit_index



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-06-22 13:11:53
Last Modified by: Hanyu Wang
Last Modified time: 2023-06-22 23:31:33
"""

from typing import List

from .base import BasicGate, MultiControlledGate, QBit, QGateType, RotationGate


MCRY_CNOT_COST = {
    "0": 0,
    "1": 2,
    "2": 4,
    "3": 8,
    "4": 16,
    "5": 32,
    "6": 64,
    "7": 128,
    "8": 226,
    "9": 290,
    "10": 362,
    "11": 442,
    "12": 530,
    "13": 626,
    "14": 730,
    "15": 842,
    "16": 962,
    "17": 1090,
    "18": 1226,
    "19": 1370,
    "20": 1522,
    "21": 1682,
    "22": 1850,
    "23": 2026,
    "24": 2210,
    "25": 2402,
    "26": 2602,
    "27": 2810,
    "28": 3026,
}


class MCRY(RotationGate, BasicGate, MultiControlledGate):
    """Classmethod to create a multi-controlled  gate .

    :param RotationGate: [description]
    :type RotationGate: [type]
    :param BasicGate: [description]
    :type BasicGate: [type]
    :param MultiControlledGate: [description]
    :type MultiControlledGate: [type]
    """

    def __init__(
        self,
        theta: float,
        control_qubits: List[QBit],
        phases: List[int],
        target_qubit: QBit,
    ) -> None:
        BasicGate.__init__(self, QGateType.MCRY, target_qubit)
        RotationGate.__init__(self, theta)
        MultiControlledGate.__init__(self, control_qubits, phases)

    def __str__(self) -> str:
        control_str = "+".join(
            [
                str(qubit) + f"[{phase}]"
                for qubit, phase in zip(self.control_qubits, self.phases)
            ]
        )
        return f"MCRY({self.theta:0.02f}, {control_str})"

    def get_cnot_cost(self) -> int:
        """Returns the cost of the cost of the gate.

        :return: [description]
        :rtype: int
        """

        index_str = str(len(self.control_qubits))
        if index_str not in MCRY_CNOT_COST:
            raise ValueError(
                f"len(self.control_qubits) = {len(self.control_qubits)} is not supported"
            )
        return MCRY_CNOT_COST[index_str]



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-08-31 13:20:14
Last Modified by: Hanyu Wang
Last Modified time: 2023-08-31 14:11:44
"""

import numpy as np


from xyz.boolean import TruthTable
from xyz.boolean.sum_of_product import sop_to_str, convert_tt_to_sop, sop_to_tt
from xyz.boolean.truth_table import (
    TruthTableEntry,
    const0_truth_table,
    const1_truth_table,
)
from xyz.circuit import MCRY, CX, RY, decompose_mcry, control_sequence_to_gates
from xyz.qstate import QState


def _get_flip_truth_table(
    state: QState,
    target_qubit: int,
    control_qubit: int,
    phase: bool,
    verbose: bool = False,
):
    """Get flip truth table .

    :param state: [description]
    :type state: QState
    :param target_qubit: [description]
    :type target_qubit: int
    :param control_qubit: [description]
    :type control_qubit: int
    :param phase: [description]
    :type phase: bool
    """
    truth_table = TruthTable(state.num_qubits)

    index_to_thetas = {}

    if verbose:
        print(
            f"target_qubit = {target_qubit}, control_qubit = {control_qubit}, phase = {phase}"
        )

    visited_index = set()
    for index in state.index_set:
        if index in visited_index:
            continue

        reversed_index = index ^ (1 << target_qubit)

        # we need to make sure that we have not visited the reversed index
        visited_index.add(index)
        visited_index.add(reversed_index)

        target_val = index >> target_qubit & 1
        final_val = (index >> control_qubit & 1) ^ phase ^ 1

        # for each index, we need to know what is the final index
        final_index = index & ~(1 << target_qubit)
        if final_val:
            final_index |= 1 << target_qubit

        if verbose:
            print(f"|{index:0{state.num_qubits}b}>: {target_val} -> {final_val}")

        # if to flip is 1, then we need to flip the target qubit
        # else if to flip is 0, then we need to make sure the target qubit is not flipped
        to_flip = target_val ^ final_val

        # the problem is simple if we already have one-to-one mapping
        if reversed_index not in state.index_set:
            truth_table.set_bit(reversed_index, to_flip)
            truth_table.set_bit(index, to_flip)

        else:
            # now we need to check if there is a conflict
            reversed_target_val = reversed_index >> target_qubit & 1
            reversed_final_val = (reversed_index >> control_qubit & 1) ^ phase ^ 1

            # if to flip is 1, then we need to flip the target qubit
            # else if to flip is 0, then we need to make sure the target qubit is not flipped
            reversed_to_flip = reversed_target_val ^ reversed_final_val
            assert reversed_to_flip != to_flip  # this should always be true

            index0 = index if not to_flip else reversed_index
            index1 = index if to_flip else reversed_index
            theta = 2 * np.arccos(
                np.sqrt(
                    state.index_to_weight[index0]
                    / (state.index_to_weight[index1] + state.index_to_weight[index0])
                )
            )

            index_to_thetas[index] = theta
            index_to_thetas[reversed_index] = theta

    return truth_table, index_to_thetas


def _reduce_qubit(state: QState, target_qubit: int, control_qubit: int, phase: bool):
    """Perform a qubit reduction on the circuit .

    :param circuit: [description]
    :type circuit: [type]
    :param state: [description]
    :type state: QState
    :param target_qubit: [description]
    :type target_qubit: int
    """

    truth_table, index_to_thetas = _get_flip_truth_table(
        state, target_qubit, control_qubit, phase
    )

    supports = state.get_supports()
    supports.remove(target_qubit)

    esop = None
    if truth_table == const0_truth_table(state.num_qubits):
        esop = []
    elif truth_table == const1_truth_table(state.num_qubits):
        raise ValueError("Cannot convert truth table to SOP")
    else:
        esop = convert_tt_to_sop(truth_table, supports)

    if esop is None:
        print(f"truth_table = {truth_table}")
        raise ValueError("Cannot convert truth table to SOP")

    # we evaluate the cost
    cost: int = 0
    for term in esop:
        term_cost = 1 << (len(term) - 1)
        cost += term_cost

    # then let us evaluate the rotation cost
    if len(index_to_thetas) > 0:
        truth_table_real = sop_to_tt(esop, state.num_qubits)

        # now we check the other index

        for index in index_to_thetas:
            if truth_table_real.get_bit(index) == TruthTableEntry.ONE:
                index_to_thetas[index] += np.pi

        for index in state.index_set:
            if index not in index_to_thetas:
                index_to_thetas[index] = 0
        supports = _get_rotation_supports(
            index_to_thetas, state.num_qubits, target_qubit
        )
        cost += 1 << (len(supports))
    return esop, index_to_thetas, cost


def _get_rotation_supports(index_to_theta: dict, num_lits: int, target_qubit: int):
    lit_to_val = {}
    supports = set()
    for index, _ in index_to_theta.items():
        for lit in range(num_lits):
            if lit == target_qubit:
                continue
            lit_val = (index >> lit) & 1
            if lit not in lit_to_val:
                lit_to_val[lit] = lit_val
            elif lit_to_val[lit] != lit_val:
                supports.add(lit)

    return supports


def _qubit_reduction_impl(
    circuit, state: QState, optimality_level: int = 1, verbose: bool = False
):
    """Reduce the number of qubit in a circuit .

    :param circuit: [description]
    :type circuit: [type]
    :param qubit_map: [description]
    :type qubit_map: [type]
    :return: [description]
    :rtype: [type]
    """

    # make greedy decision
    best_esop = None
    best_cost = None
    best_phase = None
    best_target_qubit = None
    best_control_qubit = None
    best_index_to_thetas = None
    supports = state.get_supports()

    search_done = False
    for qubit in supports:
        if search_done:
            break
        for control_qubit in supports:
            if search_done:
                break
            if qubit == control_qubit:
                continue
            for phase in [0, 1]:
                try:
                    esop, index_to_thetas, cost = _reduce_qubit(
                        state, qubit, control_qubit, phase
                    )
                    if best_esop is None or cost < best_cost:
                        best_esop = esop
                        best_cost = cost
                        best_target_qubit = qubit
                        best_control_qubit = control_qubit
                        best_phase = phase
                        best_index_to_thetas = index_to_thetas

                        if optimality_level <= 1:
                            search_done = True
                            break
                except ValueError:
                    continue

    assert best_esop is not None

    # debug
    if verbose:
        print(
            f"best_cost = {best_cost}, best_target_qubit = {best_target_qubit}, best_control_qubit = {best_control_qubit}, best_phase = {best_phase}, best_esop = {sop_to_str(best_esop)}"
        )

    # now we need to add gates
    gates = []

    # the first gate is a CX using final qubit to control the target qubit
    assert best_target_qubit != best_control_qubit
    gate = CX(
        circuit.qubit_at(best_control_qubit),
        best_phase,
        circuit.qubit_at(best_target_qubit),
    )
    gates.append(gate)

    # then depending on the rotation table, we need to add rotation gates
    if len(best_index_to_thetas) != 0:
        supports = list(
            _get_rotation_supports(
                best_index_to_thetas, state.num_qubits, best_target_qubit
            )
        )
        if len(supports) == 0:
            # we assert index_to_thetas has only one element
            assert len(best_index_to_thetas) == 1

            # get the theta
            theta = list(best_index_to_thetas.values())[0]

            # then we add a RY gate
            gates.append(RY(theta, circuit.qubit_at(best_target_qubit)))

        else:
            rotation_table = [0 for i in range(1 << len(supports))]
            for index, theta in best_index_to_thetas.items():
                # we need to find the rotation index
                rotation_index = 0
                for i, lit in enumerate(supports):
                    rotation_index |= ((index >> lit) & 1) << i
                rotation_table[rotation_index] = theta

            support_qubits = [circuit.qubit_at(lit) for lit in supports]
            target_qubit = circuit.qubit_at(best_target_qubit)
            assert best_target_qubit not in supports
            control_sequence = decompose_mcry(rotation_table)
            rotation_gates = control_sequence_to_gates(
                control_sequence, support_qubits, target_qubit
            )
            for gate in rotation_gates:
                gates.append(gate)

    target_qubit = circuit.qubit_at(best_target_qubit)
    for term in best_esop:
        # we need to decide which gate to add
        if len(term) == 1:
            lit = term[0]
            assert lit.lit != best_target_qubit
            control_qubit = circuit.qubit_at(lit.lit)
            control_phase = lit.phase
            # we need to add an X gate
            gate = CX(control_qubit, control_phase, target_qubit)
            gates.append(gate)

        else:
            # we need to add a Toffoli gate
            control_qubits = []
            control_phases = []
            for lit in term:
                assert lit.lit != best_target_qubit
                control_qubits.append(circuit.qubit_at(lit.lit))
                control_phases.append(lit.phase)
            gate = MCRY(np.pi, control_qubits, control_phases, target_qubit)
            gates.append(gate)

    # now we check the new state
    index_to_weight = {}
    for index, weight in state.index_to_weight.items():
        new_index = index & ~(1 << best_target_qubit)

        # final_val = (index >> best_control_qubit & 1) ^ best_phase ^ 1
        # new_index = new_index | (final_val << best_target_qubit)

        if new_index not in index_to_weight:
            index_to_weight[new_index] = 0

        index_to_weight[new_index] += weight

    new_state = QState(index_to_weight, state.num_qubits)

    return gates, new_state


def qubit_reduction(circuit, target_state: QState, optimality_level: int = 1):
    """Qubit reduction algorithm based on ESOP decomposition of a qubit.

    :param circuit: [description]
    :type circuit: [type]
    :param target_state: [description]
    :type target_state: QState
    :param optimality_level: [description], defaults to 1
    :type optimality_level: int, optional
    :return: [description]
    :rtype: [type]
    """

    all_gates = []
    # we stop searching for optimal solution if the number of supports is less than 5
    # we first try to qubit reduction
    num_supports = len(target_state.get_supports())
    while num_supports > 4:
        gates, new_state = _qubit_reduction_impl(
            circuit, target_state, optimality_level
        )
        for gate in gates[::-1]:
            all_gates.append(gate)
        target_state = new_state
        _num_supports = len(target_state.get_supports())
        assert _num_supports < num_supports
        num_supports = _num_supports

    return all_gates



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-09-03 21:17:33
Last Modified by: Hanyu Wang
Last Modified time: 2023-09-03 21:18:41
"""

from .network import Network



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-02-28 07:43:53
Last Modified by: Hanyu Wang
Last Modified time: 2023-03-11 20:53:23
"""

#
# reference: https://realpython.com/python-with-statement/#writing-a-sample-class-based-context-manager
#
import time

from .colors import print_green


class stopwatch:
    """Create a new Stopwatch class .

    :return: [description]
    :rtype: [type]
    """

    # name the functions run inside this context
    def __init__(self, name: str, verbose: bool = False):
        """
        @brief Initialize a : class : ` Stopwatch `. This is the constructor for : class : ` Stopwatch `.
        @param name The name of the stopwatch. Must be at least 25 characters
        """
        assert len(name) <= 25 and "stopwatch name is too long"
        self.name = name
        self.tic = None
        self.toc = None
        self.verbose = verbose

    def __enter__(self):
        """
        @brief Called when the thread enters. Stores the time in self. tic and returns it to __
        """
        self.tic = time.perf_counter()
        return self

    # reference: https://realpython.com/python-timer/#your-first-python-timer
    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        @brief Called when the process exits. Prints information about the process to the console. This is a no - op in this case.
        @param exc_type The type of exception that was raised.
        @param exc_val The value of the exception that was raised.
        @param exc_tb The traceback of the exception that was raised
        """
        self.toc = time.perf_counter()
        duration = self.toc - self.tic
        if self.name is not None and self.verbose:
            print_green(f"{self.name:<25}: {duration:>8.02f} sec")

    def time(self):
        """
        @brief Time since last call to start (). This is useful for measuring how long we've been in the middle of a test to run.
        @return The number of seconds since the start () call that took place in the test's run () method
        """
        return float(time.perf_counter() - self.tic)


GLOBAL_STOPWATCHES = {}


class GlobalStopwatch:
    """A class decorator that creates a global stopwatch .

    :return: [description]
    :rtype: [type]
    """

    # name the functions run inside this context
    def __init__(self, name: str):
        """
        @brief Initialize a : class : ` Stopwatch `. This is the constructor for : class : ` Stopwatch `.
        @param name The name of the stopwatch. Must be at least 25 characters
        """
        assert len(name) <= 25 and "stopwatch name is too long"
        self.name = name
        self.tic = None
        self.toc = None

    def __enter__(self):
        """
        @brief Called when the thread enters. Stores the time in self. tic and returns it to __
        """
        self.tic = time.perf_counter()

    # reference: https://realpython.com/python-timer/#your-first-python-timer
    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        @brief Called when the process exits. Prints information about the process to the console. This is a no - op in this case.
        @param exc_type The type of exception that was raised.
        @param exc_val The value of the exception that was raised.
        @param exc_tb The traceback of the exception that was raised
        """
        self.toc = time.perf_counter()
        if self.name not in GLOBAL_STOPWATCHES:
            GLOBAL_STOPWATCHES[self.name] = 0
        GLOBAL_STOPWATCHES[self.name] += self.toc - self.tic
        # print_green("{:<25}: {:>8.02f} sec".format(self.name, self.toc - self.tic))

    def time(self):
        """
        @brief Time since last call to start (). This is useful for measuring how long we've been in the middle of a test to run.
        @return The number of seconds since the start () call that took place in the test's run () method
        """
        return GLOBAL_STOPWATCHES[self.name]


def call_with_global_timer(func):
    """Call function with global timer .

    :param func: [description]
    :type func: [type]
    """

    def timed_func(*args, **kwargs):
        with GlobalStopwatch(func.__name__):
            return func(*args, **kwargs)

    return timed_func


def get_time(name: str):
    """Get the current stop time for a given name .

    :param name: [description]
    :type name: str
    :return: [description]
    :rtype: [type]
    """
    if name not in GLOBAL_STOPWATCHES:
        return 0
    return GLOBAL_STOPWATCHES[name]


def global_stopwatch_report():
    """Report the global stopwatch ."""
    for name, current_time in GLOBAL_STOPWATCHES.items():
        print_green(f"{name:<25}: {current_time:>8.02f} sec")



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-06-22 14:14:19
Last Modified by: Hanyu Wang
Last Modified time: 2023-06-22 15:01:37
"""

# pylint: disable=C0103

from qiskit.circuit.library.standard_gates import RYGate
from qiskit.extensions import UnitaryGate

from ..basic_gates import U, CU, MCRY


class SpecialGates:
    """SpecialGates class ."""

    def __init__(self):
        pass

    @staticmethod
    def mcry(gate: MCRY):
        """Convert a MCRY gate into a RYZ gate .

        :param gate: [description]
        :type gate: MCRY
        :return: [description]
        :rtype: [type]
        """
        num_control_qubits = len(gate.control_qubits)

        # we need to reverse the order of the control qubits
        # because the qiskit's control gate is in the reverse order
        control_str = "".join([str(int(phase)) for phase in gate.phases])[::-1]

        return RYGate(gate.get_theta()).control(
            num_control_qubits, ctrl_state=control_str
        )

    @staticmethod
    def u(gate: U):
        """Creates a CU gate .

        :param gate: [description]
        :type gate: U
        :return: [description]
        :rtype: [type]
        """

        return UnitaryGate(gate.get_unitary())

    @staticmethod
    def cu(gate: CU):
        """Creates a CU gate .

        :param gate: [description]
        :type gate: CU
        """

        num_control_qubits = 1

        # we need to reverse the order of the control qubits
        # because the qiskit's control gate is in the reverse order
        control_str = str(int(gate.phase))

        return UnitaryGate(gate.get_unitary()).control(
            num_control_qubits, ctrl_state=control_str
        )



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-11-21 22:42:21
Last Modified by: Hanyu Wang
Last Modified time: 2023-11-22 01:42:29
"""
from typing import Tuple
from collections import Counter

import copy
import numpy as np

from xyz.qstate import QState
import xyz.circuit as qc

from .library import Library


def rotation(curr_state_pair: Tuple, circuit: qc.QCircuit):
    """AI is creating summary for rotation

    :param curr_state: [description]
    :type curr_state: QState
    :param next_state: [description]
    :type next_state: QState
    :param qubit: [description]
    :type qubit: int
    :param theta: [description]
    :type theta: float
    """

    curr_cost, curr_state = curr_state_pair
    supports = curr_state.get_supports()

    for qubit_index in supports:
        thetas = curr_state.get_ry_angles(qubit_index)
        most_frequent_theta, frequency = Counter(thetas).most_common(1)[0]
        if not np.isclose(
            most_frequent_theta, 0, atol=1e-2
        ) and frequency > thetas.count(0):
            # then we apply and return
            next_state = curr_state.apply_ry(qubit_index, -most_frequent_theta)

            gates = []
            gates.append(qc.RY(most_frequent_theta, circuit.qubit_at(qubit_index)))
            return (curr_cost, next_state), gates[:]

    return None, []


def reflection(curr_state_pair: Tuple, circuit: qc.QCircuit):
    """Perform reflection of the current circuit .

    :param curr_state_pair: [description]
    :type curr_state_pair: Tuple
    :param circuit: [description]
    :type circuit: qc.QCircuit
    :return: [description]
    :rtype: [type]
    """

    curr_cost: int
    curr_state: QState

    curr_cost, curr_state = curr_state_pair
    supports = curr_state.get_supports()

    for qubit_index in supports:
        thetas = curr_state.get_ry_angles(qubit_index)
        initial_score = len(thetas) - thetas.count(0)

        rotation_table = curr_state.get_rotation_table(qubit_index)
        rotation_table_str = "\n".join(
            [
                f"{bin(idx)[2:].zfill(curr_state.num_qubits)}: {theta}"
                for idx, theta in rotation_table.items()
            ]
        )
        print(f"rotation table: \n{rotation_table_str}\n\n")

        for control_qubit_index in supports:
            if control_qubit_index == qubit_index:
                continue

            cry_thetas = curr_state.get_cry_angles(control_qubit_index, qubit_index)

            for cry_theta in cry_thetas:
                candidate_state: QState = copy.deepcopy(curr_state)
                candidate_state = candidate_state.apply_ry(qubit_index, cry_theta)
                candidate_state = candidate_state.apply_cx(
                    control_qubit_index, 0, qubit_index
                )
                candidate_state = candidate_state.apply_ry(qubit_index, -cry_theta)

                thetas = candidate_state.get_ry_angles(qubit_index)

                # if not np.isclose(
                #     most_frequent_theta, 0, atol=1e-2
                # ) and frequency > thetas.count(0):
                #     candidate_state = candidate_state.apply_ry(
                #         qubit_index, -most_frequent_theta
                #     )

                thetas = candidate_state.get_ry_angles(qubit_index)
                final_score = len(thetas) - thetas.count(0)
                curr_score = initial_score - final_score

                # this is tricky
                # if we only accept the state with lower cost, then we will not be able to reach the ground state
                # if curr_score < 0:
                if curr_score == 0:
                    continue

                # we add the gates
                gates = []
                # gates.append(qc.RY(most_frequent_theta, circuit.qubit_at(qubit_index)))
                gates.append(qc.RY(cry_theta, circuit.qubit_at(qubit_index)))
                gates.append(
                    qc.CX(
                        circuit.qubit_at(control_qubit_index),
                        0,
                        circuit.qubit_at(qubit_index),
                    )
                )
                gates.append(qc.RY(-cry_theta, circuit.qubit_at(qubit_index)))

                yield (curr_cost + 1, candidate_state), gates[:]

    return None, []


def cx(curr_state_pair: Tuple, circuit: qc.QCircuit):
    """Perform CX of the current circuit .

    :param curr_state_pair: [description]
    :type curr_state_pair: Tuple
    :param circuit: [description]
    :type circuit: qc.QCircuit
    :return: [description]
    :rtype: [type]
    """

    curr_cost, curr_state = curr_state_pair
    supports = curr_state.get_supports()

    for control_qubit_index in supports:
        for target_qubit_index in supports:
            if control_qubit_index == target_qubit_index:
                continue

            candidate_state: QState = copy.deepcopy(curr_state)
            candidate_state = candidate_state.apply_cx(
                control_qubit_index, 0, target_qubit_index
            )

            gates = []
            gates.append(
                qc.CX(
                    circuit.qubit_at(control_qubit_index),
                    0,
                    circuit.qubit_at(target_qubit_index),
                )
            )

            yield (curr_cost + 1, candidate_state), gates[:]

    return None, []


class DefaultLibrary(Library):
    """A default library class .

    :param Library: [description]
    :type Library: [type]
    """

    def explore(self):
        n_visited: int = 0

        while not self.state_queue.empty():
            curr_state_pair = self.state_queue.get()

            curr_cost: int
            curr_state: QState
            curr_cost, curr_state = curr_state_pair

            curr_repr = curr_state.repr()
            self.visited_states.add(curr_repr)
            self.record[curr_state] = curr_cost

            n_visited += 1
            if n_visited % 100 == 0:
                print(
                    f"#enquened states: {len(self.enquened_states)}, #visited states: {len(self.visited_states)}, current cost: {curr_cost}"
                )

            if curr_state == QState.ground_state(self.num_qubits):
                break

            next_state_pair, gates = rotation(curr_state_pair, self.circuit)
            if next_state_pair is not None:
                self.explore_state(curr_state_pair, next_state_pair, gates)
                continue

            for next_state_pair, gates in reflection(curr_state_pair, self.circuit):
                if next_state_pair is not None:
                    self.explore_state(curr_state_pair, next_state_pair, gates)
                    continue

            # for next_state_pair, gates in cx(curr_state_pair, self.circuit):
            #     if next_state_pair is not None:
            #         self.explore_state(curr_state_pair, next_state_pair, gates)
            #         continue

        return



import numpy as np

from xyz import QCircuit, RY, CRY, simulate_circuit, quantize_state

if __name__ == "__main__":
    circ = QCircuit(3)
    circ.add_gate(RY(2 * np.arccos(np.sqrt(2) / np.sqrt(12)), circ.qubit_at(0)))
    circ.add_gate(RY(np.pi / 2, circ.qubit_at(1)))
    circ.add_gate(
        CRY(
            2 * np.arccos(np.sqrt(6) / np.sqrt(10)),
            circ.qubit_at(0),
            1,
            circ.qubit_at(2),
        )
    )
    circ.add_gate(
        CRY(-2 * np.arccos(1 / np.sqrt(4)), circ.qubit_at(1), 1, circ.qubit_at(0))
    )

    state_vector_act = simulate_circuit(circ).data
    print("actual state: ", quantize_state(state_vector_act))
    circuit = circ.to_qiskit()
    print(circuit)



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-03-09 16:39:58
Last Modified by: Hanyu Wang
Last Modified time: 2023-03-19 00:41:28
"""

from xyz.network import Network


def write_blif_to_string(ntk: Network) -> str:
    """
    write the blif to a string
    """

    blif_string = "# Generated by MADBuf\n"

    if ntk.get_name() is None or ntk.get_name() == "":
        ntk.set_name("unknown")

    blif_string += ".model " + ntk.get_name() + "\n"

    blif_string += ".inputs "

    for input_name in ntk.pis():
        blif_string += input_name + " "

    blif_string += "\n"

    blif_string += ".outputs "

    for output in ntk.pos():
        blif_string += output + " "

    blif_string += "\n"

    for register_output, register_input in ntk.ro_to_ris().items():
        blif_string += ".latch\t" + register_input + " " + register_output + " 2\n"

    for node in ntk.constant0s():
        blif_string += ".names\t" + node + "\n0\n"

    for node in ntk.constant1s():
        blif_string += ".names\t" + node + "\n1\n"

    for node in ntk.get_nodes():
        blif_string += ".names\t"

        for fanin in ntk.fanins(node):
            blif_string += fanin + " "

        blif_string += node + "\n"

        for truth_table in ntk.get_func(node):
            blif_string += truth_table + "\n"

    blif_string += ".end\n"

    return blif_string


def write_blif(ntk: Network, filename: str):
    """
    write the blif to a file
        TODO:
            - consider also the blackboxes
    """

    file = open(filename, "w", encoding="utf-8")
    file.write(write_blif_to_string(ntk))
    file.close()



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-08-27 00:10:55
Last Modified by: Hanyu Wang
Last Modified time: 2023-08-27 00:41:20
"""

from typing import Any

from xyz.qstate import QState

from .operator import QOperatorBase, QOperatorType
from .rotation import QuantizedRotation, QuantizedRotationType
from .control import ControlledOperator


class CXOperator(QOperatorBase, QuantizedRotation, ControlledOperator):
    """Creates an CXOperator class with CXOperator .

    :param QOperatorBase: [description]
    :type QOperatorBase: [type]
    :param QuantizedRotation: [description]
    :type QuantizedRotation: [type]
    """

    def __init__(
        self,
        target_qubit_index: int,
        control_qubit_index: int,
        control_qubit_phase: bool,
    ) -> None:
        QOperatorBase.__init__(self, QOperatorType.CX, target_qubit_index)
        QuantizedRotation.__init__(self, QuantizedRotationType.SWAP)
        ControlledOperator.__init__(self, control_qubit_index, control_qubit_phase)

    def __call__(self, qstate: QState) -> Any:
        """
        This method is used to apply a quantum gate operation to a quantum state.
        @param qstate - The input quantum state.
        @return The modified quantum state after applying the gate operation.
        """
        return qstate.apply_cx(
            self.control_qubit_index, self.control_qubit_phase, self.target_qubit_index
        )

    def __str__(self) -> str:
        """The repr of the class .

        :return: [description]
        :rtype: str
        """
        phase_str = "~" if self.control_qubit_phase else ""

        return f"C({phase_str}{self.control_qubit_index})X({self.target_qubit_index})"

    def get_cost(self) -> int:
        """Returns the cost of the job .

        :return: [description]
        :rtype: int
        """
        return 1



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-09-03 21:25:18
Last Modified by: Hanyu Wang
Last Modified time: 2023-09-05 15:12:29
"""

from .write_blif import *



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-06-20 00:05:16
Last Modified by: Hanyu Wang
Last Modified time: 2023-06-20 00:06:35
"""

import numpy as np


def D_state(num_qubits: int, num_bits: int) -> np.ndarray:
    """Return a D state corresponding to the D state .

    :param num_qubits: [description]
    :type num_qubits: int
    :param num_bits: [description]
    :type num_bits: int
    :return: [description]
    :rtype: np.ndarray
    """
    state = np.zeros(2**num_qubits)

    ones: float = 0

    for i in range(2**num_qubits):
        num_ones = bin(i).count("1")
        if num_ones == num_bits:
            state[i] = 1
            ones += 1

    return 1 / np.sqrt(ones) * state



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-06-22 23:55:57
Last Modified by: Hanyu Wang
Last Modified time: 2023-06-23 00:13:28
"""

from .base import AdvancedGate, QBit, QGateType


class MULTIPLEXY(AdvancedGate):
    """Returns a DanceXRULTI class .

    :param AdvancedGate: [description]
    :type AdvancedGate: [type]
    """

    def __init__(
        self, theta0: float, theta1: float, control_qubit: QBit, target_qubit: QBit
    ) -> None:
        AdvancedGate.__init__(self, QGateType.MULTIPLEX_Y)

        assert isinstance(target_qubit, QBit)
        assert isinstance(control_qubit, QBit)

        self.target_qubit: QBit = target_qubit
        self.control_qubit: QBit = control_qubit
        self.theta0: float = theta0
        self.theta1: float = theta1

    def __str__(self) -> str:
        return f"MULTIPLEXY({self.theta0}, {self.theta1}, {self.control_qubit}, {self.target_qubit})"

    def get_cnot_cost(self) -> int:
        """Returns the cost of the cost of the gate.

        :return: [description]
        :rtype: int
        """
        return 2



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-08-27 00:21:52
Last Modified by: Hanyu Wang
Last Modified time: 2023-08-27 01:10:43
"""

import numpy as np

from typing import Any

from xyz.qstate import QState

from .operator import QOperatorBase, QOperatorType
from .rotation import QuantizedRotation, QuantizedRotationType
from .control import ControlledOperator


class CTROperator(QOperatorBase, QuantizedRotation, ControlledOperator):
    """Creates an CXOperator class with CXOperator .

    :param QOperatorBase: [description]
    :type QOperatorBase: [type]
    :param QuantizedRotation: [description]
    :type QuantizedRotation: [type]
    """

    def __init__(
        self,
        target_qubit_index: int,
        target_phase: bool,
        control_qubit_index: int,
        control_qubit_phase: bool,
    ) -> None:
        if target_phase is True:
            QOperatorBase.__init__(self, QOperatorType.CT1, target_qubit_index)
            QuantizedRotation.__init__(self, QuantizedRotationType.MERGE1)
            ControlledOperator.__init__(self, control_qubit_index, control_qubit_phase)
        else:
            QOperatorBase.__init__(self, QOperatorType.CT0, target_qubit_index)
            QuantizedRotation.__init__(self, QuantizedRotationType.MERGE0)
            ControlledOperator.__init__(self, control_qubit_index, control_qubit_phase)

        # The angle of the rotation.
        self.theta = None

    def __call__(self, qstate: QState) -> Any:
        """Apply the gate to the gate .

        :param qstate: [description]
        :type qstate: QState
        :return: [description]
        :rtype: Any
        """
        if self.operator_type == QOperatorType.CT0:
            next_state, theta = qstate.apply_controlled_merge0(
                self.control_qubit_index,
                self.control_qubit_phase,
                self.target_qubit_index,
            )
            self.theta = theta
            return next_state

        if self.operator_type == QOperatorType.CT1:
            next_state, theta = qstate.apply_controlled_merge1(
                self.control_qubit_index,
                self.control_qubit_phase,
                self.target_qubit_index,
            )

            # this is an interesting case
            self.theta = theta - np.pi
            return next_state

    def __str__(self) -> str:
        """The repr of the class .

        :return: [description]
        :rtype: str
        """
        phase_str = "~" if self.control_qubit_phase else ""
        target_phase_str = "1" if self.operator_type == QOperatorType.CT1 else "0"

        return f"C({phase_str}{self.control_qubit_index})TR({self.target_qubit_index}->{target_phase_str})"

    def get_cost(self) -> int:
        """Returns the cost of the job .

        :return: [description]
        :rtype: int
        """
        return 2



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-06-28 11:21:22
Last Modified by: Hanyu Wang
Last Modified time: 2023-06-28 11:23:03
"""

from .mcry import *
from .x import *
from .cx import *
from .operator import *
from .tr import *
from .ctr import *



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-06-28 11:36:46
Last Modified by: Hanyu Wang
Last Modified time: 2023-06-28 11:42:30
"""

from typing import Any, List

from xyz.qstate import QState

from .operator import QOperatorBase, QOperatorType
from .rotation import QuantizedRotation, QuantizedRotationType
from .control import MultiControlledOperator


class MCRYOperator(QOperatorBase, QuantizedRotation, MultiControlledOperator):
    """Construct a multi - controlledOperator .

    :param QOperatorBase: [description]
    :type QOperatorBase: [type]
    :param QuantizedRotation: [description]
    :type QuantizedRotation: [type]
    :param MultiControlledOperator: [description]
    :type MultiControlledOperator: [type]
    """

    def __init__(
        self,
        target_qubit_index: int,
        rotation_type: QuantizedRotationType,
        control_qubit_indices: List[int],
        control_qubit_phases: List[bool],
    ) -> None:
        QOperatorBase.__init__(self, QOperatorType.MCRY, target_qubit_index)
        QuantizedRotation.__init__(self, rotation_type)
        MultiControlledOperator.__init__(
            self, control_qubit_indices, control_qubit_phases
        )

    def __call__(self, qstate: QState) -> Any:
        """
        This method is used to apply a quantum gate operation to a quantum state.
        @param qstate - The input quantum state.
        @return The modified quantum state after applying the gate operation.
        """

        if self.rotation_type == QuantizedRotationType.SWAP:
            if len(self.control_qubit_indices) == 0:
                return qstate.apply_x(self.target_qubit_index)
            else:
                assert len(self.control_qubit_indices) == 1
                return qstate.apply_cx(
                    self.control_qubit_indices[0],
                    self.control_qubit_phases[0],
                    self.target_qubit_index,
                )

        if self.rotation_type == QuantizedRotationType.MERGE0:
            if len(self.control_qubit_indices) == 0:
                return qstate.apply_merge0(self.target_qubit_index)
            else:
                assert len(self.control_qubit_indices) == 1
                return qstate.apply_controlled_merge0(
                    self.control_qubit_indices[0],
                    self.control_qubit_phases[0],
                    self.target_qubit_index,
                )

        if self.rotation_type == QuantizedRotationType.MERGE1:
            assert len(self.control_qubit_indices) == 1
            return qstate.apply_controlled_merge1(
                self.control_qubit_indices[0],
                self.control_qubit_phases[0],
                self.target_qubit_index,
            )

        return qstate

    def get_cost(self) -> int:
        """Returns the cost of the gate .

        :return: [description]
        :rtype: int
        """
        num_controls: int = len(self.control_qubit_indices)
        if num_controls == 0:
            return 0
        elif num_controls == 1 and self.rotation_type == QuantizedRotationType.SWAP:
            return 1
        else:
            return 1 << (num_controls)

    def __str__(self) -> str:
        rotation_str = (
            "X"
            if self.rotation_type is QuantizedRotationType.SWAP
            else "Y"
            if self.rotation_type is QuantizedRotationType.MERGE0
            else "Y"
        )
        if len(self.control_qubit_indices) == 0:
            return f"{rotation_str}({self.target_qubit_index})"
        if len(self.control_qubit_indices) == 1:
            return f"C{rotation_str}({self.control_qubit_indices[0]}, {self.target_qubit_index})"

        return f"MCRY({self.target_qubit_index}, {rotation_str}, {self.control_qubit_indices}, {self.control_qubit_phases})"

    def __invert__(self):
        match self.rotation_type:
            case QuantizedRotationType.SWAP:
                self.rotation_type = QuantizedRotationType.SWAP

            case QuantizedRotationType.MERGE0:
                self.rotation_type = QuantizedRotationType.SPLIT0

            case QuantizedRotationType.MERGE1:
                self.rotation_type = QuantizedRotationType.SPLIT1

            case QuantizedRotationType.SPLIT0:
                self.rotation_type = QuantizedRotationType.MERGE0

            case QuantizedRotationType.SPLIT1:
                self.rotation_type = QuantizedRotationType.MERGE1

        return self



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-02-28 07:43:53
Last Modified by: Hanyu Wang
Last Modified time: 2023-03-11 20:52:12
"""

# text colors
# reference: https://www.codeproject.com/Articles/5329247/How-to-change-text-color-in-a-Linux-terminal


def print_green(text: str, **kwargs):
    """Print a green green .

    :param text: [description]
    :type text: str
    """
    print(f"\033[92m{text}\033[39m", **kwargs)


def print_red(text: str, **kwargs):
    """Print a red message with red .

    :param text: [description]
    :type text: str
    """
    print(f"\033[91m{text}\033[39m", **kwargs)


def print_yellow(text: str, **kwargs):
    """Print a yellow color.

    :param text: [description]
    :type text: str
    """
    print(f"\033[93m{text}\033[39m", **kwargs)


def text_red(text: str):
    """Return a text representation of text .

    :param text: [description]
    :type text: str
    :return: [description]
    :rtype: [type]
    """
    return f"\033[91m{text}\033[39m"


def text_yellow(text: str):
    """Return a pretty printed text for text .

    :param text: [description]
    :type text: str
    :return: [description]
    :rtype: [type]
    """
    return f"\033[93m{text}\033[39m"


def print_blue(text: str, **kwargs):
    """Print a blue color .

    :param text: [description]
    :type text: str
    """
    print(f"\033[94m{text}\033[39m", **kwargs)


def print_orange(text: str, **kwargs):
    """Print an Orange color .

    :param text: [description]
    :type text: str
    """
    print(f"\033[93m{text}\033[39m", **kwargs)


def text_orange(text: str):
    return f"\033[93m{text}\033[39m"



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-06-28 11:36:46
Last Modified by: Hanyu Wang
Last Modified time: 2023-06-28 11:42:30
"""

from typing import Any

from xyz.qstate import QState

from .operator import QOperatorBase, QOperatorType
from .rotation import QuantizedRotation, QuantizedRotationType


class XOperator(QOperatorBase, QuantizedRotation):
    """Creates an XOperator class with XOperator .

    :param QOperatorBase: [description]
    :type QOperatorBase: [type]
    :param QuantizedRotation: [description]
    :type QuantizedRotation: [type]
    """

    def __init__(
        self,
        target_qubit_index: int,
    ) -> None:
        QOperatorBase.__init__(self, QOperatorType.X, target_qubit_index)
        QuantizedRotation.__init__(self, QuantizedRotationType.SWAP)

    def __call__(self, qstate: QState) -> Any:
        """
        This method is used to apply a quantum gate operation to a quantum state.
        @param qstate - The input quantum state.
        @return The modified quantum state after applying the gate operation.
        """
        return qstate.apply_x(self.target_qubit_index)

    def __str__(self) -> str:
        """The repr of the class .

        :return: [description]
        :rtype: str
        """

        return f"X({self.target_qubit_index})"

    def get_cost(self) -> int:
        """Returns the cost of the job .

        :return: [description]
        :rtype: int
        """
        return 0



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-09-03 11:15:09
Last Modified by: Hanyu Wang
Last Modified time: 2023-09-06 10:33:15
"""


class Network:
    """A class method for creating a network ."""

    def __init__(self):
        self.__top_module = ""
        self.__inputs = set()
        self.__outputs = set()
        self.__nodes = set()
        self.__register_inputs = set()
        self.__register_outputs = set()
        self.__ro_to_ri: dict = {}

        # __signals is a list of all the __nodes in the network in the topological order
        # this is private and should not be modified directly
        self.__signals = []

        self.__const0 = set()
        self.__const1 = set()

        # node fanins return the set of fanins of a node
        #  - note that only __nodes can be looked up in this dictionary
        #  - __signals are not safe when directly looked up
        self.__node_fanins: dict = {}
        self.__node_funcs: dict = {}

        self.__node_fanouts: dict = {}

    def get_name(self):
        """Returns name of the module .

        :return: [description]
        :rtype: [type]
        """
        return self.__top_module

    def set_name(self, name: str):
        """AI is creating summary for set_name

        :param name: [description]
        :type name: str
        """
        self.__top_module = name

    def get_func(self, node: str):
        """Returns the function func for the given node

        :param node: [description]
        :type node: str
        :return: [description]
        :rtype: [type]
        """
        return self.__node_funcs[node]

    def ro_to_ris(self):
        """Returns a dictionary mapping from register_output to

        :return: [description]
        :rtype: [type]
        """
        return self.__ro_to_ri

    def is_po(self, signal: str) -> bool:
        """Returns True if signal is a POINT .

        :param signal: [description]
        :type signal: str
        :return: [description]
        :rtype: bool
        """
        return signal in self.__outputs

    def is_pi(self, signal: str) -> bool:
        """Returns True if signal is a valid signal .

        :param signal: [description]
        :type signal: str
        :return: [description]
        :rtype: bool
        """
        return signal in self.__inputs

    def is_ro(self, signal: str) -> bool:
        """Returns True if signal is a RPC signal .

        :param signal: [description]
        :type signal: str
        :return: [description]
        :rtype: bool
        """
        return signal in self.__register_outputs

    def is_ri(self, signal: str) -> bool:
        """Returns True if signal is a RRI .

        :param signal: [description]
        :type signal: str
        :return: [description]
        :rtype: bool
        """
        return signal in self.__register_inputs

    def num_nodes(self) -> int:
        """Returns the number of nodes in the graph .

        :return: [description]
        :rtype: int
        """
        return len(self.__nodes)

    def num_latch(self) -> int:
        """AI is creating summary for num_latch

        :return: [description]
        :rtype: int
        """
        return len(self.__register_outputs)

    def num_pis(self) -> int:
        """Returns the number of pis in the network .

        :return: [description]
        :rtype: int
        """
        return len(self.__inputs)

    def num_pos(self) -> int:
        """The number of outputs in the model .

        :return: [description]
        :rtype: int
        """
        return len(self.__outputs)

    # the CO (combinational __outputs are the primary __outputs and the register inupts)
    def is_co(self, signal: str) -> bool:
        """Returns True if signal is a co - op code .

        :param signal: [description]
        :type signal: str
        :return: [description]
        :rtype: bool
        """
        return signal in self.__outputs or signal in self.__register_inputs

    # the CI (combinational inptus are the primary __inputs and the register __outputs)
    def is_ci(self, signal: str) -> bool:
        """Returns True if signal is a combinational input .

        :param signal: [description]
        :type signal: str
        :return: [description]
        :rtype: bool
        """
        return (
            signal in self.__inputs
            or signal in self.__register_outputs
            or signal in self.__const0
            or signal in self.__const1
        )

    def topological_traversal(self) -> set:
        """The topological traversal set of signals in the graph .

        :return: [description]
        :rtype: set
        """
        return self.__signals

    def constants(self):
        """Returns a list of all constant constants .

        :return: [description]
        :rtype: [type]
        """
        return sorted(self.__const0 | self.__const1)

    def cos(self):
        """The list of all outputs .

        :return: [description]
        :rtype: [type]
        """
        return sorted(self.__outputs | self.__register_inputs)

    def cis(self):
        """Returns a list of all the combinational inputs.

        :return: [description]
        :rtype: [type]
        """
        return sorted(
            self.__inputs | self.__register_outputs | self.__const0 | self.__const1
        )

    def constant0s(self):
        """Returns a list of constant 0s .

        :return: [description]
        :rtype: [type]
        """
        return sorted(self.__const0)

    def constant1s(self):
        """The list of constant 1s .

        :return: [description]
        :rtype: [type]
        """
        return sorted(self.__const1)

    def pis(self) -> list:
        """The list of pis in the order of the PIs .

        :return: [description]
        :rtype: [type]
        """
        return sorted(self.__inputs)

    def pos(self):
        """Returns a sorted list of all outputs in the order they are sorted .

        :return: [description]
        :rtype: [type]
        """
        return sorted(self.__outputs)

    def ris(self):
        """The list of register inputs sorted by the receiver .

        :return: [description]
        :rtype: [type]
        """
        return sorted(self.__register_inputs)

    def ros(self):
        """The list of RRS outputs sorted by register outputs .

        :return: [description]
        :rtype: [type]
        """
        return sorted(self.__register_outputs)

    def fanins(self, signal: str):
        """Returns a list of the fanins of a specified signal .

        :param signal: [description]
        :type signal: str
        :return: [description]
        :rtype: [type]
        """
        return sorted(self.__node_fanins[signal])

    def get_nodes(self):
        """Get the list of nodes sorted by their indices .

        :return: [description]
        :rtype: [type]
        """
        return sorted(self.__nodes)

    # sort __signals in a topological order
    def traverse(self):
        """
        Traverse the CIS .
            TODO: support runtime modification and maintain the topogical order
        """
        self.__signals = []
        for signal in self.cis():
            assert signal not in self.__signals
            self.__signals.append(signal)

        for signal in self.cos():
            self.trav_rec(signal)

        for signal in self.__signals:
            self.__node_fanouts[signal] = set()

        # prepare fanouts: this should be recomputed after each network modification
        for signal in self.__signals:
            if signal in self.__node_fanins:
                for fanin_signal in self.fanins(signal):
                    self.__node_fanouts[fanin_signal].add(signal)

    # topological traversal, used to sort the __signals in a topological order
    def trav_rec(self, signal: str, pending_signals: set = None):
        """recursive recursion recursively

        :param signal: [description]
        :type signal: str
        :param pending_signals: [description], defaults to set()
        :type pending_signals: set, optional
        """
        if signal in self.__signals:
            return

        if signal not in self.__node_fanins:
            print(f"recursion stoped at node {signal}")
            exit()

        pending_signals.add(signal)

        for f in self.fanins(signal):
            assert f != signal, f"node {signal} is its own fanin"
            if f not in self.__signals:
                if f in pending_signals:
                    # we have a loop
                    # print(f"recursion stoped at node {signal}")
                    # print(f"pending signals: {pending_signals}")
                    return
                self.trav_rec(f)

        pending_signals.remove(signal)
        self.__signals.append(signal)

    def num_fanouts(self, signal: str):
        """Returns the number of fanouts in the node .

        :param signal: [description]
        :type signal: str
        :return: [description]
        :rtype: [type]
        """
        return len(self.__node_fanouts[signal])

    #
    # graph modifications
    #
    def create_pi(self, name: str):
        """Create a input with the given name .

        :param name: [description]
        :type name: str
        """
        assert name not in self.__inputs and "the input to create already exists"
        self.__inputs.add(name)

    def create_po(self, name: str):
        """Create a new output .

        :param name: [description]
        :type name: str
        """
        assert name not in self.__outputs and "the output to create already exists"
        self.__outputs.add(name)

    def create_ri(self, name: str):
        """Create a new RRI input

        :param name: [description]
        :type name: str
        """
        assert (
            name not in self.__register_inputs
            and "the register input to create already exists"
        )
        self.__register_inputs.add(name)

    def create_ro(self, name: str):
        """Create a new register output with the given name .

        :param name: [description]
        :type name: str
        """
        assert (
            name not in self.__register_outputs
            and "the register output to create already exists"
        )
        self.__register_outputs.add(name)

    def create_node(self, name: str, fanins: set, func: list):
        """Create a new node with the given name, fanins and function .

        :param name: [description]
        :type name: str
        :param fanins: [description]
        :type fanins: set
        :param func: [description]
        :type func: list
        """
        assert name not in self.__nodes and "the node to create already exists"
        self.__nodes.add(name)
        self.__node_fanins[name] = set(list(fanins)[:])  # deep copy
        self.__node_funcs[name] = func[:]  # deep copy
        self.__node_fanouts[name] = set()

    def create_and(self, f1: str, f2: str, name: str):
        """Create and connect a node with f1 and f2 .

        :param f1: [description]
        :type f1: str
        :param f2: [description]
        :type f2: str
        :param name: [description]
        :type name: str
        """
        self.create_node(name=name, fanins=set([f1, f2]), func=["11 1"])

    def create_or(self, f1: str, f2: str, name: str):
        """Create or update a node with a or - 1 .

        :param f1: [description]
        :type f1: str
        :param f2: [description]
        :type f2: str
        :param name: [description]
        :type name: str
        """
        self.create_node(name=name, fanins=set([f1, f2]), func=["1- 1", "-1 1"])

    def create_buf(self, fin: str, fout: str):
        """create a buffer for writing fout

        :param fin: [description]
        :type fin: str
        :param fout: [description]
        :type fout: str
        """
        self.create_node(name=fout, fanins=set([fin]), func=["1 1"])

    def create_latch(self, register_input: str, register_output: str):
        """Creates a Latch with the given RAP and R .

        :param register_input: [description]
        :type register_input: str
        :param register_output: [description]
        :type register_output: str
        """
        self.__register_inputs.add(register_input)
        self.__register_outputs.add(register_output)
        self.__ro_to_ri[register_output] = register_input



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-08-31 14:07:02
Last Modified by: Hanyu Wang
Last Modified time: 2023-08-31 16:45:33
"""

from typing import List
from queue import PriorityQueue

from itertools import chain, combinations


from .truth_table import (
    TruthTable,
    const1_truth_table,
    create_truth_table,
    const0_truth_table,
)


class Lit:
    """Class method to create a literal ."""

    def __init__(self, lit, phase: bool) -> None:
        self.lit = lit
        self.phase = phase

    def __str__(self) -> str:
        return f"{'~' if not self.phase else ''}{self.lit}"

    def __eq__(self, __value: object) -> bool:
        return self.lit == __value.lit and self.phase == __value.phase

    def __hash__(self) -> int:
        return hash(self.lit << 1 | self.phase)

    def to_value(self):
        """Convert to a bit value .

        :return: [description]
        :rtype: [type]
        """
        return self.lit << 1 | self.phase

    def __lt__(self, other):
        return self.to_value() < other.to_value()

    def __le__(self, other):
        return self.to_value() <= other.to_value()


def powerset(iterable):
    """Poweret of an iterable .

    powerset([1,2,3]) --> () (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)
    :param iterable: [description]
    :type iterable: [type]
    :return: [description]
    :rtype: [type]
    """
    elements = list(iterable)
    return chain.from_iterable(
        combinations(elements, r) for r in range(len(elements) + 1)
    )


def term_to_str(term: List[Lit]):
    """Convert a list of terms to a string .

    :param term: [description]
    :type term: List[Lit]
    :return: [description]
    :rtype: [type]
    """
    return "*".join([str(lit) for lit in term])


def sop_to_str(sop: List[List[Lit]]):
    """Convert a list of expressions into a string .

    :param sop: [description]
    :type sop: List[List[Lit]]
    :return: [description]
    :rtype: [type]
    """
    return " ^ ".join([term_to_str(term) for term in sop])


def convert_tt_to_sop(
    truth_table: TruthTable,
    supports: List[int] = None,
    use_esop: bool = True,
    optimality_level: int = 1,
):
    """Convert a truth table to a sum of product .

    :param truth_table: [description]
    :type truth_table: TruthTable
    :return: [description]
    :rtype: [type]
    """

    num_lits = len(truth_table)

    all_lits = range(num_lits) if supports is None else supports

    lit_to_tt = {}
    unate_terms = []
    binate_terms = []

    # preprocess some unate terms from the binate terms
    if not use_esop:
        for lit in all_lits:
            tt_lit_pos = create_truth_table(num_lits, lit)
            tt_lit_neg = ~tt_lit_pos
            lit_to_tt[Lit(lit, True)] = tt_lit_pos
            lit_to_tt[Lit(lit, False)] = tt_lit_neg

            if tt_lit_pos < truth_table:
                unate_terms.append([Lit(lit, True)])

            if tt_lit_neg < truth_table:
                unate_terms.append([Lit(lit, False)])

            if not tt_lit_pos < ~truth_table:
                binate_terms.append(Lit(lit, True))

            if not tt_lit_neg < ~truth_table:
                binate_terms.append(Lit(lit, False))
    # if we use ESOP, we need to add all the binate terms
    else:
        for lit in all_lits:
            tt_lit_pos = create_truth_table(num_lits, lit)
            tt_lit_neg = ~tt_lit_pos
            lit_to_tt[Lit(lit, True)] = tt_lit_pos
            lit_to_tt[Lit(lit, False)] = tt_lit_neg

            binate_terms.append(Lit(lit, True))
            binate_terms.append(Lit(lit, False))

    # we can do a DFS here to find all the unate terms
    def dfs(curr_tt: TruthTable, curr_term: List[Lit]):
        """Perform a DFS to find all the unate terms .

        :param curr_tt: [description]
        :type curr_tt: TruthTable
        :param curr_term: [description]
        :type curr_term: List[Lit]
        """
        nonlocal unate_terms
        nonlocal truth_table

        if curr_tt == const0_truth_table(num_lits):
            return

        if len(curr_term) > 0:
            unate_terms.append(curr_term[:])

        # keep adding the literals would not make the truth table smaller
        if curr_tt < truth_table:
            return

        for lit in binate_terms:
            if len(curr_term) > 0 and lit <= curr_term[-1]:
                continue
            next_tt = curr_tt & lit_to_tt[lit]
            curr_term.append(lit)
            dfs(next_tt, curr_term)
            curr_term.pop()

    dfs(const1_truth_table(num_lits), [])

    if optimality_level <= 1:
        sum_terms = []
        unate_terms = sorted(unate_terms)
        curr_tt = truth_table
        while curr_tt != const0_truth_table(num_lits):
            for term in unate_terms:
                term_tt = sop_to_tt([term], num_lits, use_esop)

                if term_tt < curr_tt:
                    curr_tt = curr_tt ^ term_tt
                    sum_terms.append(term)
        return sum_terms

    # apply a BFS
    q = PriorityQueue()
    enqueued = {}
    prev = {}

    def visit(new_cost: int, new_tt: TruthTable, curr_tt: TruthTable, lit: Lit):
        """Visit a new truth table .

        :param new_cost: [description]
        :type new_cost: int
        :param new_tt: [description]
        :type new_tt: TruthTable
        """
        if str(new_tt) in enqueued and enqueued[str(new_tt)] <= new_cost:
            return False
        enqueued[str(new_tt)] = new_cost
        prev[str(new_tt)] = (curr_tt, lit)
        q.put((new_cost, new_tt))
        return True

    # record the current cost and the current truth table
    curr_tt = None

    visit(0, const0_truth_table(num_lits), None, None)

    while not q.empty():
        curr_cost, curr_tt = q.get()

        if curr_tt == truth_table:
            break

        for term in unate_terms:
            # we skip the empty term
            if len(term) == 0:
                continue
            term_cost = 1 << (len(term) - 1)

            term_tt = None
            for lit in term:
                if term_tt is None:
                    term_tt = lit_to_tt[lit]
                else:
                    term_tt = term_tt & lit_to_tt[lit]

            if not term_tt < curr_tt:
                next_cost = curr_cost + term_cost
                if use_esop:
                    next_tt = curr_tt ^ term_tt
                else:
                    next_tt = curr_tt | term_tt
                visit(next_cost, next_tt, curr_tt, term)

    if not curr_tt == truth_table:
        return None

    # we reconstruct the sum of product
    sum_terms = []

    prev_tt = curr_tt
    while prev_tt is not None:
        prev_tt, term = prev[str(prev_tt)]

        if prev_tt is None:
            break

        sum_terms.append(term)

    # we return a list of list
    return sum_terms


def sop_to_tt(sop: List[List[Lit]], num_lits: int, use_esop: bool = True):
    """Convert a sum of product to a truth table .

    :param sop: [description]
    :type sop: List[List[Lit]]
    :return: [description]
    :rtype: [type]
    """
    truth_table = const0_truth_table(num_lits)
    for term in sop:
        term_tt = const1_truth_table(num_lits)
        for lit in term:
            if lit.phase:
                term_tt = term_tt & create_truth_table(num_lits, lit.lit)
            else:
                term_tt = term_tt & ~create_truth_table(num_lits, lit.lit)
        if use_esop:
            truth_table = truth_table ^ term_tt
        else:
            truth_table = truth_table | term_tt
    return truth_table



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-08-12 03:02:33
Last Modified by: Hanyu Wang
Last Modified time: 2023-08-19 13:40:18
"""

import json

from typing import List, Tuple
import numpy as np

# the merge uncertainty, if the difference between the two angles is less than
# this value, we consider them to be the same
MERGE_UNCERTAINTY = 1e-3


N_DIGITS = 2


class QState:
    """Class method for QState"""

    from ._rotation_angles import (
        get_cry_angles,
        get_ry_angles,
        get_most_frequent_theta,
        get_rotation_table,
    )

    def __init__(self, index_to_weight: dict, num_qubit: int) -> None:
        self.num_qubits = num_qubit

        self.index_to_weight = {}
        for index, weight in index_to_weight.items():
            if not np.isclose(weight, 0, atol=1e-3):
                self.index_to_weight[index] = weight
        self.index_set = self.index_to_weight.keys()

        self.sparsity: int = len(index_to_weight)

    def __deepcopy__(self, memo):
        return QState(self.index_to_weight, self.num_qubits)

    def apply_x(self, qubit_index: int) -> "QState":
        """Apply X gate to the qubit.

        :param qubit_index: [description]
        :type qubit_index: int
        """
        index_to_weight = {}
        for idx in self.index_set:
            reversed_idx = idx ^ (1 << qubit_index)
            index_to_weight[reversed_idx] = self.index_to_weight[idx]
        return QState(index_to_weight, self.num_qubits)

    def apply_cx(
        self, control_qubit_index: int, phase: bool, target_qubit_index: int
    ) -> "QState":
        """Apply CX gate to the qubit.

        :param control_qubit_index: [description]
        :type control_qubit_index: int
        :param target_qubit_index: [description]
        :type target_qubit_index: int
        """
        index_to_weight = {}
        for idx, weight in self.index_to_weight.items():
            reversed_idx = idx ^ (1 << target_qubit_index)
            if (idx >> control_qubit_index) & 1 == phase:
                index_to_weight[reversed_idx] = weight
            else:
                index_to_weight[idx] = weight
        return QState(index_to_weight, self.num_qubits)

    def apply_ry(self, qubit_index: int, theta: float) -> "QState":
        """Apply the Y operator to the given qubit .

        :param qubit_index: [description]
        :type qubit_index: int
        """
        index_to_weight = {idx: 0 for idx in self.index_set}
        for idx, weight in self.index_to_weight.items():
            rdx = idx ^ (1 << qubit_index)
            if rdx not in self.index_set:
                index_to_weight[rdx] = 0

            if idx >> qubit_index & 1 == 0:
                index_to_weight[idx] += weight * np.cos(theta / 2)
                index_to_weight[rdx] += weight * np.sin(theta / 2)
            else:
                index_to_weight[idx] += weight * np.cos(theta / 2)
                index_to_weight[rdx] -= weight * np.sin(theta / 2)
        return QState(index_to_weight, self.num_qubits)

    def apply_merge0(self, qubit_index: int) -> "QState":
        """Apply the Y operator to the given qubit .

        :param qubit_index: [description]
        :type qubit_index: int
        """
        index_to_weight = {}
        theta = None
        for idx in self.index_set:
            reversed_idx = idx ^ (1 << qubit_index)
            if reversed_idx not in self.index_set:
                raise ValueError("The state is not a valid state.")
            idx0 = idx & ~(1 << qubit_index)
            idx1 = idx0 ^ (1 << qubit_index)

            # now we check the rotation angle
            weight0 = self.index_to_weight[idx0]
            weight1 = self.index_to_weight[idx1]
            weight_total = np.sqrt(
                (weight0 ** 2) + (weight1 ** 2)
            )

            _theta = 2 * np.arctan(weight1 / weight0)

            if theta is None:
                theta = _theta
            elif not np.isclose(theta, _theta, atol=MERGE_UNCERTAINTY):
                raise ValueError("The state is not a valid state.")

            # finally we update the weight
            index_to_weight[idx0] = weight_total
        if theta is None:
            raise ValueError("The state is not a valid state.")
        return QState(index_to_weight, self.num_qubits), theta

    def apply_merge1(self, qubit_index: int) -> "QState":
        """Apply the Y operator to the given qubit .

        :param qubit_index: [description]
        :type qubit_index: int
        """
        index_to_weight = {}
        theta = None
        for idx in self.index_set:
            reversed_idx = idx ^ (1 << qubit_index)
            if reversed_idx not in self.index_set:
                raise ValueError("The state is not a valid state.")
            idx0 = idx & ~(1 << qubit_index)
            idx1 = idx0 ^ (1 << qubit_index)

            # now we check the rotation angle
            weight0 = self.index_to_weight[idx0]
            weight1 = self.index_to_weight[idx1]
            weight_total = np.sqrt(
                (weight0 ** 2) + (weight1 ** 2)
            )

            _theta = 2 * np.arctan(weight0 / weight1)

            if theta is None:
                theta = _theta
            elif not np.isclose(theta, _theta, atol=MERGE_UNCERTAINTY):
                raise ValueError("The state is not a valid state.")

            # finally we update the weight
            index_to_weight[idx1] = weight_total
        if theta is None:
            raise ValueError("The state is not a valid state.")
        return QState(index_to_weight, self.num_qubits), theta

    def apply_split(self, qubit_index: int) -> None:
        """Apply the Y operator to the given qubit .

        :param qubit_index: [description]
        :type qubit_index: int
        """
        new_states = []
        for state in self.index_set:
            state0 = state & ~(1 << qubit_index)
            state1 = state | (1 << qubit_index)

            new_states.append(state0)
            new_states.append(state1)

        next_state = QState(new_states, self.num_qubits)
        return next_state

    def apply_controlled_merge0(
        self, control_qubit_index: int, phase: bool, target_qubit_index: int
    ) -> None:
        """Apply the Y operator to the given qubit .

        :param qubit_index: [description]
        :type qubit_index: int
        """
        index_to_weight = {}
        theta = None
        for idx in self.index_set:
            # no rotation
            if (idx >> control_qubit_index) & 1 != phase:
                index_to_weight[idx] = self.index_to_weight[idx]
                continue

            reversed_idx = idx ^ (1 << target_qubit_index)
            if reversed_idx not in self.index_set:
                raise ValueError("The state is not a valid state.")
            idx0 = idx & ~(1 << target_qubit_index)
            idx1 = idx0 ^ (1 << target_qubit_index)

            # now we check the rotation angle
            weight0 = self.index_to_weight[idx0]
            weight1 = self.index_to_weight[idx1]
            weight_total = np.sqrt(
                (weight0 ** 2) + (weight1 ** 2)
            )

            _theta = 2 * np.arctan(weight1 / weight0)
            
            if theta is None:
                theta = _theta
            elif not np.isclose(theta, _theta, atol=MERGE_UNCERTAINTY):
                raise ValueError("The state is not a valid state.")

            # finally we update the weight
            index_to_weight[idx0] = weight_total
        if theta is None:
            raise ValueError("The state is not a valid state.")
        return QState(index_to_weight, self.num_qubits), theta

    def apply_controlled_merge1(
        self, control_qubit_index: int, phase: bool, target_qubit_index: int
    ) -> None:
        """Apply the Y operator to the given qubit .

        :param qubit_index: [description]
        :type qubit_index: int
        """
        index_to_weight = {}
        theta = None
        for idx in self.index_set:
            # no rotation
            if (idx >> control_qubit_index) & 1 != phase:
                index_to_weight[idx] = self.index_to_weight[idx]
                continue

            reversed_idx = idx ^ (1 << target_qubit_index)
            if reversed_idx not in self.index_set:
                raise ValueError("The state is not a valid state.")
            idx0 = idx & ~(1 << target_qubit_index)
            idx1 = idx0 ^ (1 << target_qubit_index)

            # now we check the rotation angle
            weight0 = self.index_to_weight[idx0]
            weight1 = self.index_to_weight[idx1]
            weight_total = np.sqrt(
                (weight0 ** 2) + (weight1 ** 2)
            )

            _theta = 2 * np.arctan(weight1 / weight0)

            if theta is None:
                theta = _theta
            elif not np.isclose(theta, _theta, atol=MERGE_UNCERTAINTY):
                raise ValueError("The state is not a valid state.")

            # finally we update the weight
            index_to_weight[idx1] = weight_total
        if theta is None:
            raise ValueError("The state is not a valid state.")
        return QState(index_to_weight, self.num_qubits), theta

    def get_supports(self) -> List[int]:
        """Return the support of the state .

        :param state: [description]
        :type state: QState
        :return: [description]
        :rtype: List[int]
        """
        signatures = self.get_qubit_signatures()
        qubit_indices = []
        for qubit, pattern in enumerate(signatures):
            if pattern != 0:
                qubit_indices.append(qubit)
        return qubit_indices

    def get_sparsity(self) -> int:
        """Return the sparsity of the state .

        :param state: [description]
        :type state: QState
        :return: [description]
        :rtype: int
        """
        return len(self.index_set)

    def to_value(self) -> int:
        """Return the value of the state .

        :return: [description]
        :rtype: int
        """
        value = 0
        for basis in self.index_set:
            value |= 1 << basis
        return value

    def get_qubit_signatures(self) -> List[int]:
        """Transpose the state array ."""
        signatures = [0 for i in range(self.num_qubits)]
        for _, value in enumerate(self.index_set):
            for j in range(self.num_qubits):
                signatures[j] = signatures[j] << 1 | (value >> j & 1)
        return signatures

    def get_const1_signature(self) -> int:
        """Returns the number of signed unsigned signatures .

        :return: [description]
        :rtype: int
        """
        return (1 << len(self.index_set)) - 1

    def cofactors(self, pivot_qubit: int) -> Tuple["QState", "QState"]:
        """Returns the cofactors of the given qubit .

        :return: [description]
        :rtype: [type]
        """

        index_to_weight0 = {}
        index_to_weight1 = {}

        total_weights0 = 0
        total_weights1 = 0

        for idx in self.index_set:
            if (idx >> pivot_qubit) & 1 == 0:
                index_to_weight0[idx] = self.index_to_weight[idx]
                total_weights0 += self.index_to_weight[idx]
            else:
                index_to_weight1[idx ^ (1 << pivot_qubit)] = self.index_to_weight[idx]
                total_weights1 += self.index_to_weight[idx]

        return (
            QState(index_to_weight0, self.num_qubits),
            QState(index_to_weight1, self.num_qubits),
            total_weights0,
            total_weights1,
        )

    @staticmethod
    def ground_state(num_qubits: int) -> "QState":
        """Return the ground state .

        :param num_qubits: [description]
        :type num_qubits: int
        :return: [description]
        :rtype: QState
        """
        state = QState({0: 1.0}, num_qubits)
        return state

    def get_lower_bound(self) -> int:
        """Returns the lower bound of the state .

        :param state: [description]
        :type state: QState
        :return: [description]
        :rtype: int
        """
        lower_bound: int = 0
        signatures = self.get_qubit_signatures()
        for pattern in signatures:
            if pattern != 0:
                lower_bound += 1
        return lower_bound

    def __str__(self) -> str:
        return " + ".join(
            [
                f"{weight.real:0.0{N_DIGITS}f}*|{idx:0{self.num_qubits}b}>".zfill(
                    self.num_qubits
                )
                for idx, weight in self.index_to_weight.items()
            ]
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, QState):
            return False
        return self.__hash__() == other.__hash__()

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, QState):
            return False
        sorted_index_set = sorted(self.index_set, reverse=True)
        sorted_o_index_set = sorted(other.index_set, reverse=True)
        for i, index in enumerate(sorted_index_set):
            if i >= len(sorted_o_index_set):
                return False
            if index < sorted_o_index_set[i]:
                return True
            elif index > sorted_o_index_set[i]:
                return False
        return False

    def __hash__(self) -> int:
        return hash(tuple(sorted(self.index_set)))
        # return hash(str(self))

    def repr(self) -> int:
        """Return a hex representation of the bitmap .

        :return: [description]
        :rtype: int
        """
        # self.index_set = sorted(self.index_set)
        # signatures = self.get_qubit_signatures()
        # return hash(tuple(sorted(signatures, key=lambda x: bin(x).count("1"))))
        return hash(self)

    def to_vector(self) -> np.ndarray:
        """Return the vector representation of the state .

        :return: [description]
        :rtype: np.ndarray
        """
        vector = np.zeros(2**self.num_qubits)
        for idx, weight in self.index_to_weight.items():
            vector[idx] = np.sqrt(weight)

        # normalize the vector
        vector /= np.linalg.norm(vector)

        return vector

    def to_file(self, filename: str):
        """Writes the benchmark to a file .

        :param filename: [description]
        :type filename: str
        """

        # we format the index as a string
        index_to_weight = {
            f"{index:0{self.num_qubits}b}": weight
            for index, weight in self.index_to_weight.items()
        }
        benchmark_str = json.dumps(index_to_weight)
        # lets be safe here

        with open(filename, "w", encoding="utf-8") as file:
            file.write(benchmark_str)


def quantize_state(state_vector: np.ndarray):
    """Quantize a state to the number of qubits .

    :param state_vector: a vector with 2**n entries, where n is the number of qubits.
    :type state_vector: np.ndarray
    """

    # discard the imaginary part
    # state_vector = state_vector.real
    state_vector = np.real(state_vector)

    # normalize the vector
    state_vector = state_vector.astype(np.float64) / np.linalg.norm(
        state_vector.astype(np.float64)
    )

    index_to_weight = {}
    num_qubits = int(np.log2(len(state_vector)))
    for idx, coefficient in enumerate(state_vector):
        if not np.isclose(coefficient, 0):
            index_to_weight[idx] = coefficient
    return QState(index_to_weight, num_qubits)


def load_state(filename: str):
    """Loads the state of a given file .

    :param filename: [description]
    :type filename: str
    """

    # read the dict from the file
    with open(filename, "r", encoding="utf-8") as file:
        state_dict = json.load(file)

    num_qubits: int = None
    index_to_weight = {}
    for index_str, weight in state_dict.items():
        if num_qubits is None:
            num_qubits = len(index_str)

        index = int(index_str, 2)
        index_to_weight[index] = weight

    # convert the dict to a QState
    return QState(index_to_weight, num_qubits)


def from_val(val: int, num_qubits: int) -> QState:
    """Return the state from the vector representation .

    :param state_vector: [description]
    :type state_vector: np.ndarray
    :return: [description]
    :rtype: QState
    """

    assert val > 0 and val < 2 ** (2**num_qubits)
    states = []
    for i in range(2**num_qubits):
        if val & 1 == 1:
            states.append(i)
        val >>= 1

    patterns = [0 for i in range(num_qubits)]
    for _, value in enumerate(states):
        for j in range(num_qubits):
            patterns[j] = patterns[j] << 1 | ((value >> j) & 1)
    return QState(patterns, len(states))



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-08-31 14:05:07
Last Modified by: Hanyu Wang
Last Modified time: 2023-08-31 16:34:17
"""

from enum import Enum, auto
from math import log
from typing import Any


class TruthTableEntry(Enum):
    """Class method .

    :param enum: [description]
    :type enum: [type]
    """

    ZERO = auto()
    ONE = auto()
    DONT_CARE = auto()


TT_STR = {
    TruthTableEntry.ZERO: "0",
    TruthTableEntry.ONE: "1",
    TruthTableEntry.DONT_CARE: "-",
}


class TruthTable:
    """Class method ."""

    def __init__(self, num_lits: int) -> None:
        """Initialize a truth table with the given number of qubits .

        :param num_lits: [description]
        :type num_lits: int
        """
        self._num_lits = num_lits
        self._table = [TruthTableEntry.DONT_CARE for i in range(2**num_lits)]

    def __str__(self) -> str:
        return "".join([TT_STR[entry] for entry in self._table])

    def __len__(self):
        return self._num_lits

    def set_true(self, bit):
        """Set the given bit to true .

        :param bit: [description]
        :type bit: [type]
        """
        self._table[bit] = TruthTableEntry.ONE

    def set_false(self, bit):
        """Set the given bit to false .

        :param bit: [description]
        :type bit: [type]
        """
        self._table[bit] = TruthTableEntry.ZERO

    def set_bit(self, bit, value):
        """Set the given bit to the given value .

        :param bit: [description]
        :type bit: [type]
        :param value: [description]
        :type value: [type]
        """
        if value:
            self.set_true(bit)
        else:
            self.set_false(bit)

    def __lt__(self, other):
        """Compare two truth tables .

        :param other: [description]
        :type other: [type]
        :return: [description]
        :rtype: [type]
        """
        for i in range(2**self._num_lits):
            if self._table[i] == TruthTableEntry.DONT_CARE:
                continue
            if other._table[i] == TruthTableEntry.DONT_CARE:
                continue
            if (
                self._table[i] == TruthTableEntry.ONE
                and other._table[i] == TruthTableEntry.ZERO
            ):
                return False

        return True

    def __or__(self, __value: Any) -> "TruthTable":
        """Compute the disjunction of two truth tables .

        :param __value: [description]
        :type __value: Any
        :return: [description]
        :rtype: TruthTable
        """
        tt_or = TruthTable(self._num_lits)
        for i in range(2**self._num_lits):
            if (
                self._table[i] == TruthTableEntry.ONE
                or __value._table[i] == TruthTableEntry.ONE
            ):
                tt_or._table[i] = TruthTableEntry.ONE
            elif (
                self._table[i] == TruthTableEntry.ZERO
                and __value._table[i] == TruthTableEntry.ZERO
            ):
                tt_or._table[i] = TruthTableEntry.ZERO
            else:
                tt_or._table[i] = TruthTableEntry.DONT_CARE

        return tt_or

    def __and__(self, __value: Any) -> "TruthTable":
        """Compute the conjunction of two truth tables .

        :param __value: [description]
        :type __value: Any
        :return: [description]
        :rtype: TruthTable
        """
        tt_and = TruthTable(self._num_lits)
        for i in range(2**self._num_lits):
            if (
                self._table[i] == TruthTableEntry.ZERO
                or __value._table[i] == TruthTableEntry.ZERO
            ):
                tt_and._table[i] = TruthTableEntry.ZERO
            elif (
                self._table[i] == TruthTableEntry.ONE
                and __value._table[i] == TruthTableEntry.ONE
            ):
                tt_and._table[i] = TruthTableEntry.ONE
            else:
                tt_and._table[i] = TruthTableEntry.DONT_CARE

        return tt_and

    def __invert__(self):
        """Negate a truth table .

        :return: [description]
        :rtype: [type]
        """
        negated = TruthTable(self._num_lits)
        for i in range(2**self._num_lits):
            if self._table[i] == TruthTableEntry.DONT_CARE:
                continue
            if self._table[i] == TruthTableEntry.ONE:
                negated.set_false(i)
            else:
                negated.set_true(i)
        return negated

    def __add__(self, other):
        """Compute the sum of two truth tables .

        :param other: [description]
        :type other: [type]
        :return: [description]
        :rtype: [type]
        """
        sum_tt = TruthTable(self._num_lits)
        for i in range(2**self._num_lits):
            if self._table[i] == TruthTableEntry.ONE:
                sum_tt._table[i] = other._table[i]
            elif other._table[i] == TruthTableEntry.ZERO:
                sum_tt._table[i] = TruthTableEntry.ZERO
        return sum_tt

    def __sub__(self, other):
        """Compute the difference of two truth tables .

        :param other: [description]
        :type other: [type]
        :return: [description]
        :rtype: [type]
        """
        sub_tt = TruthTable(self._num_lits)
        for i in range(2**self._num_lits):
            if self._table[i] == TruthTableEntry.ONE:
                assert other._table[i] == TruthTableEntry.ONE
                sub_tt._table[i] = TruthTableEntry.ONE
            elif self._table[i] == TruthTableEntry.ZERO:
                if other._table[i] == TruthTableEntry.ONE:
                    sub_tt._table[i] = TruthTableEntry.ZERO
                else:
                    sub_tt._table[i] = TruthTableEntry.DONT_CARE
        return sub_tt

    def is_dont_care(self):
        """Check if the truth table is a don't care .

        :return: [description]
        :rtype: [type]
        """
        for entry in self._table:
            if entry != TruthTableEntry.DONT_CARE:
                return False
        return True

    def __eq__(self, __value: object) -> bool:
        for i in range(2**self._num_lits):
            if self._table[i] == TruthTableEntry.DONT_CARE:
                continue
            if __value._table[i] == TruthTableEntry.DONT_CARE:
                continue
            if self._table[i] != __value._table[i]:
                return False
        return True

    def __xor__(self, __value: object) -> bool:
        tt_xor = TruthTable(self._num_lits)
        for i in range(2**self._num_lits):
            if self._table[i] == TruthTableEntry.DONT_CARE:
                continue
            if __value._table[i] == TruthTableEntry.DONT_CARE:
                continue
            if self._table[i] != __value._table[i]:
                tt_xor._table[i] = TruthTableEntry.ONE
            else:
                tt_xor._table[i] = TruthTableEntry.ZERO
        return tt_xor

    def get_bit(self, bit):
        """Get the bit value of the given bit .

        :param bit: [description]
        :type bit: [type]
        :return: [description]
        :rtype: [type]
        """
        return self._table[bit]

    def get_bit_str(self, bit):
        """Get the bit string of the given bit .

        :param bit: [description]
        :type bit: [type]
        :return: [description]
        :rtype: [type]
        """
        return TT_STR[self._table[bit]]


def create_truth_table(num_lits: int, idx: int):
    """Create a truth table with the given number of literals and the given index .

    :param num_lits: [description]
    :type num_lits: int
    :param idx: [description]
    :type idx: int
    :return: [description]
    :rtype: [type]
    """
    truth_table = TruthTable(num_lits)

    for i in range(1 << num_lits):
        if (i >> idx) & 1:
            truth_table.set_true(i)
        else:
            truth_table.set_false(i)
    return truth_table


def const0_truth_table(num_lits: int):
    """Create a truth table with the given number of literals and the given index .

    :param num_lits: [description]
    :type num_lits: int
    :param idx: [description]
    :type idx: int
    :return: [description]
    :rtype: [type]
    """
    truth_table = TruthTable(num_lits)

    for i in range(1 << num_lits):
        truth_table.set_false(i)
    return truth_table


def const1_truth_table(num_lits: int):
    """Create a truth table with the given number of literals and the given index .

    :param num_lits: [description]
    :type num_lits: int
    :param idx: [description]
    :type idx: int
    :return: [description]
    :rtype: [type]
    """
    truth_table = TruthTable(num_lits)

    for i in range(1 << num_lits):
        truth_table.set_true(i)
    return truth_table


def read_truth_table(truth_table_str: str):
    """Reads a truth table from a truth table .

    :param truth_table_str: [description]
    :type truth_table_str: str
    """

    num_lits = int(log(len(truth_table_str), 2))

    truth_table = TruthTable(num_lits)

    for index, char in enumerate(truth_table_str):
        if char == "0":
            truth_table.set_false(index)
        elif char == "1":
            truth_table.set_true(index)

        # the other cases are don't care

    return truth_table



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-06-18 11:32:39
Last Modified by: Hanyu Wang
Last Modified time: 2023-06-18 11:32:58
"""

# https://stackoverflow.com/questions/47561840/how-can-i-separate-the-functions-of-a-class-into-multiple-files


# standard library
from typing import List

# internal modules
from .basic_gates import QGate, QGateType, QBit
from ._optimization import _add_gate_optimized, _add_gates_optimized
from ._qiskit import _to_qiskit


class QCircuit:
    """the class of quantum circuit"""

    def __init__(
        self, num_qubits: int, map_gates: bool = False, qubits: list = None
    ) -> None:
        """Initialize the gate .

        :param num_qubits: [description]
        :type num_qubits: int
        """
        self.__gates: List[QGate] = []

        if qubits is not None:
            self.__qubits: List[QBit] = qubits[:]
        else:
            self.__qubits: List[QBit] = []
            self.init_qubits(num_qubits)

        # configures
        self.map_gates = map_gates

    def sub_circuit(self, supports: list) -> "QCircuit":
        """Get the sub-circuit of the current circuit .

        :param qubit_mapping: [description]
        :type qubit_mapping: [type]
        :return: [description]
        :rtype: [type]
        """
        qubits = [self.qubit_at(i) for i in supports]
        sub_circuit = QCircuit(len(qubits), map_gates=self.map_gates, qubits=qubits)
        return sub_circuit

    def get_num_qubits(self) -> int:
        """
        Get the number of qubits in the circuit
        """
        return len(self.__qubits)

    def qubit_at(self, index: int) -> QBit:
        """
        Get the qubit at the specified index
        """
        assert (
            index < self.get_num_qubits()
        ), f"index = {index} >= num_qubits = {self.get_num_qubits()}"
        return self.__qubits[index]

    def get_gates(self) -> List[QGate]:
        """
        Get the gates in the circuit
        """
        return self.__gates

    def add_gate(self, gate: QGate):
        """
        Add a gate to the circuit
        """
        _add_gate_optimized(self, gate)

    def add_gates(self, gates: List[QGate]):
        """
        Add a list of gates to the circuit
        """
        _add_gates_optimized(self, gates)

    def append_gate(self, gate: QGate):
        """
        Add a gate to the circuit
        """
        self.__gates.append(gate)

    def append_gates(self, gates: List[QGate]):
        """
        Add a list of gates to the circuit
        """
        self.__gates.extend(gates)

    def add_qubit(self, qubit: QBit):
        """
        Add a qubit to the circuit
        """
        self.__qubits.append(qubit)

    def add_qubits(self, qubits: List[QBit]):
        """
        Add a list of qubits to the circuit
        """
        self.__qubits.extend(qubits)

    def init_qubits(self, num_qubits: int):
        """
        Add a list of qubits to the circuit
        """
        for i in range(num_qubits):
            self.add_qubit(QBit(i))

    def num_gates(self, gate_type: QGateType = None) -> int:
        """
        Get the number of gates in the circuit
        """
        if gate_type is None:
            return len(self.__gates)
        else:
            return len(
                [gate for gate in self.__gates if gate.get_qgate_type() == gate_type]
            )

    def get_cnot_cost(self) -> int:
        """
        Get the number of CNOT gates in the circuit
        """
        cnot_cost: int = 0
        for gate in self.__gates:
            cnot_cost += gate.get_cnot_cost()
        return cnot_cost

    def to_qiskit(self):
        """Convert the sequence to a Qiskit string .

        :param with_measurement: [description], defaults to True
        :type with_measurement: bool, optional
        :param with_tomography: [description], defaults to False
        :type with_tomography: bool, optional
        :return: [description]
        :rtype: [type]
        """
        return _to_qiskit(self)

    def __str__(self) -> str:
        """Convert the sequence to a string .

        :return: [description]
        :rtype: str
        """
        qubit_index_list = [qubit.index for qubit in self.__qubits]
        qubit_index_list_str = ",".join([str(i) for i in qubit_index_list])
        return f"Circuit({qubit_index_list_str})"



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-09-01 12:51:03
Last Modified by: Hanyu Wang
Last Modified time: 2023-09-01 12:56:17
"""

import copy
from collections import Counter

import xyz.qstate as qs
import xyz.circuit as qc


def heuristic_cnot_synthesis(
    circuit: qc.QCircuit,
    target_state: qs.QState,
    verbose_level: int = 0,
):
    """This function finds the exact cnot_cnot_synthesis of a circuit .

    :param circuit: [description]
    :type circuit: QCircuit
    :param qubit_mapping: [description]
    :type qubit_mapping: dict
    :param target_state: [description]
    :type target_state: qs.QState
    :param optimality_level: [description], defaults to 3
    :type optimality_level: int, optional
    :param verbose_level: [description], defaults to 0
    :type verbose_level: int, optional
    :raises ValueError: [description]
    :return: [description]
    :rtype: [type]
    """

    curr_state = copy.deepcopy(target_state)

    if verbose_level >= 3:
        print(f"preparing the target state {target_state} ...")

    gates = []

    while curr_state != qs.QState.ground_state(target_state.num_qubits):
        print(f"\t\tcurrent state: {curr_state}")

        supports = curr_state.get_supports()

        # we decide the best qubit to be the target qubit
        best_qubit_index: int = None
        best_control_qubit_index: int = None
        best_cry_theta: float = None
        best_ry_theta: float = None
        best_score: int = 0

        for qubit_index in supports:
            thetas = curr_state.get_ry_angles(qubit_index)

            # we calculate the RY score
            # print(f"qubit {qubit_index}: {thetas}")
            initial_score = len(thetas) - thetas.count(0)
            # print(f"initial score: {initial_score}, state = {rotated_state}")

            # let us first try to apply the RY gate
            most_frequent_theta, frequency = Counter(thetas).most_common(1)[0]
            print(f"most frequent theta: {most_frequent_theta}, frequency: {frequency}")
            if most_frequent_theta != 0 and frequency > thetas.count(0):
                # then we apply and return
                curr_state = curr_state.apply_ry(qubit_index, -most_frequent_theta)
                gates.append(qc.RY(most_frequent_theta, circuit.qubit_at(qubit_index)))
                break

            print(f"\n\nstate before: {curr_state} score {initial_score}")

            for control_qubit_index in supports:
                if control_qubit_index == qubit_index:
                    continue

                cry_thetas = curr_state.get_cry_angles(control_qubit_index, qubit_index)

                for cry_theta in cry_thetas:
                    candidate_state: qs.QState
                    candidate_state = curr_state.apply_ry(qubit_index, cry_theta)
                    candidate_state = candidate_state.apply_cx(
                        control_qubit_index, 0, qubit_index
                    )
                    candidate_state = candidate_state.apply_ry(qubit_index, -cry_theta)

                    thetas = candidate_state.get_ry_angles(qubit_index)
                    most_frequent_theta = Counter(thetas).most_common(1)[0][0]
                    candidate_state = candidate_state.apply_ry(
                        qubit_index, -most_frequent_theta
                    )

                    thetas = candidate_state.get_ry_angles(qubit_index)

                    final_score = len(thetas) - thetas.count(0)

                    curr_score = initial_score - final_score

                    print(f"state after: {candidate_state} score {final_score}")
                    if curr_score > best_score:
                        best_score = curr_score
                        best_qubit_index = qubit_index
                        best_control_qubit_index = control_qubit_index
                        best_cry_theta = cry_theta
                        best_ry_theta = most_frequent_theta

        if best_qubit_index is not None:
            print(
                f"best qubit: {best_qubit_index}, best control qubit: {best_control_qubit_index}, best cry theta: {best_cry_theta}"
            )

            assert best_qubit_index is not None
            assert best_cry_theta is not None
            assert best_ry_theta is not None
            best_ry_theta: float
            best_cry_theta: float

            curr_state = curr_state.apply_ry(best_qubit_index, best_cry_theta)
            curr_state = curr_state.apply_cx(
                best_control_qubit_index, 0, best_qubit_index
            )
            curr_state = curr_state.apply_ry(best_qubit_index, -best_cry_theta)
            curr_state = curr_state.apply_ry(best_qubit_index, -best_ry_theta)

            # we add the gates
            gates.append(qc.RY(-best_cry_theta, circuit.qubit_at(best_qubit_index)))
            gates.append(
                qc.CX(
                    circuit.qubit_at(best_control_qubit_index),
                    0,
                    circuit.qubit_at(best_qubit_index),
                )
            )
            gates.append(qc.RY(best_cry_theta, circuit.qubit_at(best_qubit_index)))
            gates.append(qc.RY(best_ry_theta, circuit.qubit_at(best_qubit_index)))

    return gates[::-1]



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-08-31 12:44:45
Last Modified by: Hanyu Wang
Last Modified time: 2023-08-31 12:46:26
"""

from .decompose_mcry import *



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-08-12 01:43:39
Last Modified by: Hanyu Wang
Last Modified time: 2023-08-12 01:44:38
"""

from enum import Enum, auto


class QuantizedRotationType(Enum):
    """This class method is used to instantiate a Quantized RotationType object .

    :param Enum: [description]
    :type Enum: [type]
    """

    SWAP = auto()
    MERGE0 = auto()
    MERGE1 = auto()
    SPLIT0 = auto()
    SPLIT1 = auto()


class QuantizedRotation:
    """Classmethod to convert a QuantizedRotation to QuantizedRotation ."""

    def __init__(self, rotation_type: QuantizedRotationType) -> None:
        self.rotation_type = rotation_type



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-11-21 21:50:01
Last Modified by: Hanyu Wang
Last Modified time: 2023-11-21 21:50:03
"""

from .library import Library
from .default_library import DefaultLibrary



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-08-27 00:18:14
Last Modified by: Hanyu Wang
Last Modified time: 2023-08-27 00:39:26
"""

import numpy as np

from typing import Any
from xyz.qstate import QState

from .operator import QOperatorBase, QOperatorType
from .rotation import QuantizedRotation, QuantizedRotationType


class TROperator(QOperatorBase, QuantizedRotation):
    """Creates an TROperator class with TROperator .

    :param QOperatorBase: [description]
    :type QOperatorBase: [type]
    :param QuantizedRotation: [description]
    :type QuantizedRotation: [type]
    """

    def __init__(
        self,
        target_qubit_index: int,
        target_phase: bool,
    ) -> None:
        if target_phase is True:
            QOperatorBase.__init__(self, QOperatorType.T1, target_qubit_index)
            QuantizedRotation.__init__(self, QuantizedRotationType.MERGE1)
        else:
            QOperatorBase.__init__(self, QOperatorType.T0, target_qubit_index)
            QuantizedRotation.__init__(self, QuantizedRotationType.MERGE0)

        # The angle of the rotation.
        self.theta = None

    def __call__(self, qstate: QState) -> Any:
        """
        This method is used to apply a quantum gate operation to a quantum state.
        @param qstate - The input quantum state.
        @return The modified quantum state after applying the gate operation.
        """
        if self.operator_type == QOperatorType.T0:
            next_state, theta = qstate.apply_merge0(self.target_qubit_index)
            self.theta = theta
            return next_state
        if self.operator_type == QOperatorType.T1:
            next_state, theta = qstate.apply_merge1(self.target_qubit_index)
            # this is an interesting case, we need to rotate to the opposite direction
            self.theta = theta - np.pi # we need to rotate another pi
            return next_state

    def __str__(self) -> str:
        """The repr of the class .

        :return: [description]
        :rtype: str
        """
        target_phase_str = "1" if self.operator_type == QOperatorType.T1 else "0"

        return f"TR({self.target_qubit_index}->{target_phase_str})"

    def get_cost(self) -> int:
        """Returns the cost of the job .

        :return: [description]
        :rtype: int
        """
        return 0



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-06-22 13:11:53
Last Modified by: Hanyu Wang
Last Modified time: 2023-06-22 20:14:09
"""

from .base import BasicGate, QGateType, QBit

# pylint: disable=C0103


class Z(BasicGate):
    """Classmethod to create a Gate class .

    :param BasicGate: [description]
    :type BasicGate: [type]
    """

    def __init__(self, target_qubit: QBit) -> None:
        BasicGate.__init__(self, QGateType.Z, target_qubit)

    def __str__(self) -> str:
        return f"Z({self.target_qubit})"

    def get_cnot_cost(self) -> int:
        """Returns the cost of the cost of the gate.

        :return: [description]
        :rtype: int
        """
        return 0



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-06-22 14:39:56
Last Modified by: Hanyu Wang
Last Modified time: 2023-06-22 15:09:23
"""

from typing import List

from .qubit import QBit


class ControlledGate:
    """Classmethod for creating a controlled gate ."""

    def __init__(self, control_qubit: QBit, phase: int = 1) -> None:
        self.phase = phase

        assert isinstance(control_qubit, QBit)
        self.control_qubit = control_qubit

    def get_control_qubit(self) -> QBit:
        """Returns the control qubit .

        :return: [description]
        :rtype: QBit
        """
        return self.control_qubit

    def get_phase(self) -> int:
        """Returns the phase .

        :return: [description]
        :rtype: int
        """
        return self.phase


class MultiControlledGate:
    """Class method for creating a multiControlledGate ."""

    def __init__(self, control_qubits: List[QBit], phases: List[int]) -> None:
        self.control_qubits = control_qubits
        self.phases = list(phases)

    def has_zero_controls(self) -> bool:
        """Returns True if any control qubits are zero .

        :return: [description]
        :rtype: bool
        """
        return len(self.control_qubits) == 0

    def has_one_control(self) -> bool:
        """Returns True if this instruction has one control .

        :return: [description]
        :rtype: bool
        """
        return len(self.control_qubits) == 1

    def get_control_qubits(self) -> List[QBit]:
        """Returns the control qubits .

        :return: [description]
        :rtype: List[QBit]
        """
        return self.control_qubits

    def get_phases(self) -> List[int]:
        """Returns the phases .

        :return: [description]
        :rtype: List[int]
        """
        return self.phases



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-09-16 15:01:42
Last Modified by: Hanyu Wang
Last Modified time: 2023-09-16 15:04:34
"""

from typing import List

from .base import AdvancedGate, QBit, QGateType


class MCMY(AdvancedGate):
    """Returns a multi-controlled mux Y class .

    :param AdvancedGate: [description]
    :type AdvancedGate: [type]
    """

    def __init__(
        self,
        rotation_table: List[float],
        control_qubits: List[QBit],
        target_qubit: QBit,
    ) -> None:
        AdvancedGate.__init__(self, QGateType.MCMY)

        assert isinstance(target_qubit, QBit)
        assert isinstance(control_qubits, list)

        self.target_qubit: QBit = target_qubit
        self.control_qubits: QBit = control_qubits
        self.rotation_table: List[float] = rotation_table

    def __str__(self) -> str:
        control_qubit_str = ",".join([str(qubit) for qubit in self.control_qubits])
        return f"MCMY({control_qubit_str})"

    def get_cnot_cost(self) -> int:
        """Returns the cost of the cost of the gate.

        :return: [description]
        :rtype: int
        """
        return 1 << len(self.control_qubits)



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-09-13 10:29:30
Last Modified by: Hanyu Wang
Last Modified time: 2023-09-13 10:33:49
"""

# pylint: disable=C0103

import numpy as np

from .base import BasicGate, QBit, QGateType, UnitaryGate


class U(UnitaryGate, BasicGate):
    """Constructs a Unitary gate .

    :param RotationGate: [description]
    :type RotationGate: [type]
    :param BasicGate: [description]
    :type BasicGate: [type]
    """

    def __init__(self, unitary: np.ndarray, target_qubit: QBit) -> None:
        BasicGate.__init__(self, QGateType.U, target_qubit)
        UnitaryGate.__init__(self, unitary)

    def __str__(self) -> str:
        return "U"

    def get_cnot_cost(self) -> int:
        """Returns the cost of the cost of the gate.

        :return: [description]
        :rtype: int
        """
        return 0



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-06-22 13:24:31
Last Modified by: Hanyu Wang
Last Modified time: 2023-06-22 23:39:10
"""

# standard library
from typing import List

# third party library
from qiskit import QuantumCircuit, QuantumRegister

# my own library
from .basic_gates import QBit, QGate, QGateType
from .super_gates import SpecialGates


def _to_qiskit(self) -> QuantumCircuit:
    """Convert this circuit to a QuantumCircuit .

    :param with_measurement: [description], defaults to True
    :type with_measurement: bool, optional
    :param with_tomography: [description], defaults to False
    :type with_tomography: bool, optional
    :return: [description]
    :rtype: QuantumCircuit
    """
    num_qubits = self.get_num_qubits()

    quantum_registers = QuantumRegister(num_qubits)

    circuit = QuantumCircuit(quantum_registers)

    def _to_register(qubit: QBit | List[QBit]) -> QuantumRegister:
        """Converts a single bit value to a register .

        :param qubit: [description]
        :type qubit: QBit
        :return: [description]
        :rtype: QuantumRegister
        """
        nonlocal quantum_registers
        if isinstance(qubit, QBit):
            return quantum_registers[qubit.index]
        if isinstance(qubit, list):
            return [quantum_registers[q.index] for q in qubit]

    gate: QGate
    for gate in self.get_gates():
        match gate.get_qgate_type():
            case QGateType.U:
                special_gate = SpecialGates.cu(gate)
                circuit.append(
                    special_gate,
                    _to_register([gate.target_qubit]),
                )

            case QGateType.CU:
                special_gate = SpecialGates.cu(gate)
                circuit.append(
                    special_gate,
                    _to_register([gate.control_qubit, gate.target_qubit]),
                )

            case QGateType.MCRY:
                special_gate = SpecialGates.mcry(gate)
                circuit.append(
                    special_gate,
                    _to_register(gate.control_qubits + [gate.target_qubit]),
                )

            case QGateType.CX:
                circuit.cx(
                    _to_register(gate.control_qubit),
                    _to_register(gate.target_qubit),
                    ctrl_state=gate.phase,
                )

            case QGateType.RX:
                circuit.rx(gate.theta, _to_register(gate.target_qubit))

            case QGateType.RY:
                circuit.ry(gate.theta, _to_register(gate.target_qubit))

            case QGateType.RZ:
                circuit.rz(gate.theta, _to_register(gate.target_qubit))

            case QGateType.Z:
                circuit.z(_to_register(gate.target_qubit))

            case QGateType.X:
                circuit.x(_to_register(gate.target_qubit))

            case QGateType.CRX:
                circuit.crx(
                    gate.theta,
                    _to_register(gate.control_qubit),
                    _to_register(gate.target_qubit),
                    ctrl_state=gate.phase,
                )

            case QGateType.CRY:
                circuit.cry(
                    gate.theta,
                    _to_register(gate.control_qubit),
                    _to_register(gate.target_qubit),
                    ctrl_state=gate.phase,
                )

            case QGateType.CRZ:
                circuit.crz(
                    gate.theta,
                    _to_register(gate.control_qubit),
                    _to_register(gate.target_qubit),
                    ctrl_state=gate.phase,
                )

            case _:
                raise NotImplementedError(
                    f"Gate type {gate.get_qgate_type()} is not supported yet\n Consider toggle map_gates to True"
                )

    return circuit



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-09-13 12:02:08
Last Modified by: Hanyu Wang
Last Modified time: 2023-09-13 12:04:42
"""

import numpy as np
from qiskit import Aer, transpile

from xyz.qstate import quantize_state
from xyz.qstate import index_to_str
from xyz.utils import print_green, print_yellow, print_red
from xyz.circuit import QCircuit

from .diff_states import get_difference


def simulate_circuit(circuit: QCircuit):
    """Simulate a circuit .

    :param circuit: [description]
    :type circuit: QCircuit
    """
    qiskit_circuit = circuit.to_qiskit()

    transpiled = transpile(
        qiskit_circuit, basis_gates=["u", "cx"], optimization_level=0
    )

    backend = Aer.get_backend("qasm_simulator", precision="double")
    transpiled.save_statevector()
    state_vector_actual = backend.run(transpiled).result().get_statevector()
    state_vector_actual = state_vector_actual / np.linalg.norm(state_vector_actual)

    return state_vector_actual


def verify_circuit_and_count_cnot(
    circuit: QCircuit, state_vector_expect: np.ndarray, skip_verify: bool = False
):
    """Verify that the given circuit is correct .

    :param circuit: [description]
    :type circuit: QCircuit
    :param state_vector_expect: [description]
    :type state_vector_expect: np.ndarray
    :raises ValueError: [description]
    """
    qiskit_circuit = circuit.to_qiskit()

    transpiled = transpile(
        qiskit_circuit, basis_gates=["u", "cx"], optimization_level=0
    )
    cx = transpiled.count_ops()["cx"] if "cx" in transpiled.count_ops() else 0

    if not skip_verify:
        backend = Aer.get_backend("qasm_simulator")
        transpiled.save_statevector()
        state_vector_actual = backend.run(transpiled).result().get_statevector()

        # normalize actual
        state_vector_actual = state_vector_actual / np.linalg.norm(state_vector_actual)
        state_vector_actual = abs(state_vector_actual.data)
        state_vector_expect = abs(state_vector_expect)

        if not np.allclose(state_vector_actual, state_vector_expect, atol=0.1):
            state_actual = quantize_state(state_vector_actual)
            state_expect = quantize_state(state_vector_expect)

            diff_index_1, diff_index_2, diff_weights = get_difference(
                state_actual, state_expect
            )

            print_red(index_to_str(diff_index_1, state_actual.num_qubits))
            print_green(index_to_str(diff_index_2, state_actual.num_qubits))
            print_yellow(index_to_str(diff_weights, state_actual.num_qubits))
            return cx, False

    return cx, True



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-06-28 10:36:25
Last Modified by: Hanyu Wang
Last Modified time: 2023-06-28 11:02:26
"""


import logging
from sys import stdout

from xyz.utils.colors import print_green

from xyz.circuit import QCircuit
from xyz.qstate import QState

from ._exact_cnot_synthesis import exact_cnot_synthesis
from ._qubit_reduction import qubit_reduction
from ._sparse_state_synthesis import cardinality_reduction
from ._ground_state_calibration import ground_state_calibration
from ._support_reduction import support_reduction
from ._qubit_decomposition import qubit_decomposition


def intialize_logger():
    """Initialize the logger .

    :return: [description]
    :rtype: [type]
    """
    # create logger with 'spam_application'
    log = logging.getLogger("synthesis")
    log.setLevel(logging.DEBUG)

    # create formatter and add it to the handlers
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )

    # create file handler which logs even debug messages
    file_handler = logging.FileHandler("synthesis.log")
    file_handler.setLevel(logging.INFO)
    file_handler.setFormatter(formatter)
    log.addHandler(file_handler)

    # create console handler with a higher log level
    console_handler = logging.StreamHandler()
    console_handler.setLevel(logging.ERROR)
    console_handler.setFormatter(formatter)
    log.addHandler(console_handler)

    return log


def cnot_synthesis(
    state: QState,
    optimality_level: int = 3,
    map_gates: bool = False,
    verbose_level: int = 0,
    cnot_limit: int = None,
    reduction_method: str = "qubit",
) -> QCircuit:
    """
    @brief Runs the search based state synthesis
    @param verbose_level Whether to print out the state of the search
    """

    num_qubits = state.num_qubits

    # initialize the circuit
    circuit = QCircuit(num_qubits, map_gates=map_gates)

    if reduction_method == "qubit":
        gates = qubit_decomposition(
            circuit, state, optimality_level, verbose_level, cnot_limit=cnot_limit
        )
        circuit.add_gates(gates)
        return circuit

    post_processing_gates = []
    pre_processing_gates = []

    # reduce the number of qubits
    if optimality_level <= 0:
        gates = qubit_reduction(circuit, state, optimality_level)
        for gate in gates[::-1]:
            post_processing_gates.append(gate)

    # deep copy
    curr_state = QState(state.index_to_weight, state.num_qubits)

    gates = []

    skip_exact_cnot_synthesis: bool = False

    while True:
        density = len(curr_state.index_set)

        if verbose_level >= 3:
            print(f"Current state: {curr_state}, density: {density}")

        # we reach the end of the state
        if density == 1:
            _gates = ground_state_calibration(circuit, curr_state)
            for gate in _gates:
                pre_processing_gates.append(gate)
            break

        num_supports = len(curr_state.get_supports())

        if not skip_exact_cnot_synthesis and num_supports <= 4:
            try:
                if verbose_level >= 2:
                    print(
                        f"running exact cnot synthesis with optimality_level={optimality_level}",
                        end="...",
                    )
                stdout.flush()
                _gates = exact_cnot_synthesis(
                    circuit,
                    curr_state,
                    optimality_level=optimality_level,
                    verbose_level=verbose_level,
                    cnot_limit=cnot_limit,
                )
                if verbose_level >= 2:
                    print_green("done")
                for gate in _gates:
                    gates.append(gate)

                break
            # the solution is not in the induced state transition graph
            except ValueError:
                skip_exact_cnot_synthesis = True

        if verbose_level >= 2:
            print(f"reducing density, current density = {density}", end="...")
        stdout.flush()
        new_state, _gates = cardinality_reduction(
            circuit, curr_state, verbose_level=verbose_level
        )
        if verbose_level >= 2:
            print_green("done")

        for gate in _gates:
            post_processing_gates.append(gate)

        if verbose_level >= 2:
            print(f"reducing supports, num_supports = {num_supports}", end="...")
        stdout.flush()
        support_reduced_state, support_reduction_gates = support_reduction(
            circuit, new_state
        )
        curr_state = support_reduced_state
        num_supports: int = len(curr_state.get_supports())
        if verbose_level >= 2:
            print_green(f"done, num_supports = {num_supports}")

        for gate in support_reduction_gates:
            post_processing_gates.append(gate)

    for gate in pre_processing_gates:
        circuit.add_gate(gate)

    for gate in gates:
        circuit.add_gate(gate)

    for gate in post_processing_gates[::-1]:
        circuit.add_gate(gate)

    return circuit



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-09-15 13:22:42
Last Modified by: Hanyu Wang
Last Modified time: 2023-09-15 13:30:56
"""

from xyz.qstate.qstate import QState


def get_difference(state1: QState, state2: QState):
    """Returns the difference between two states .

    :param state1: [description]
    :type state1: QState
    :param state2: [description]
    :type state2: QState
    :return: [description]
    :rtype: [type]
    """

    diff_index_1 = {}
    diff_index_2 = {}
    diff_weights = {}

    for index in state1.index_to_weight:
        if index not in state2.index_to_weight:
            diff_index_1[index] = state1.index_to_weight[index]
        elif state1.index_to_weight[index] != state2.index_to_weight[index]:
            diff: float = state1.index_to_weight[index] - state2.index_to_weight[index]
            if diff >= 1e-6:
                diff_index_1[index] = diff

    for index in state2.index_to_weight:
        if index not in state1.index_to_weight:
            diff_index_2[index] = state2.index_to_weight[index]

    return diff_index_1, diff_index_2, diff_weights



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-06-22 14:14:14
Last Modified by: Hanyu Wang
Last Modified time: 2023-06-22 14:14:28
"""

from .special_gates import *



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-06-28 11:27:44
Last Modified by: Hanyu Wang
Last Modified time: 2023-06-28 11:29:03
"""

from typing import List


class ControlledOperator:
    """Class method for creating the ControlledOperator ."""

    def __init__(self, control_qubit_index: int, control_qubit_phase: bool) -> None:
        self.control_qubit_index: int = control_qubit_index
        self.control_qubit_phase: bool = control_qubit_phase

    def is_controlled(self, basis_state: int) -> bool:
        """Returns whether the basis state is controlled by the basis state .

        :param basis_state: [description]
        :type basis_state: int
        :return: [description]
        :rtype: bool
        """
        return (basis_state >> self.control_qubit_index) & 1 == self.control_qubit_phase


class MultiControlledOperator:
    """Class method to create a class that is used by the FirmrolledOperator ."""

    def __init__(
        self, control_qubit_indices: List[int], control_qubit_phases: List[bool]
    ) -> None:
        self.control_qubit_indices = control_qubit_indices
        self.control_qubit_phases = control_qubit_phases

    def is_controlled(self, basis_state: int) -> bool:
        """Returns whether the basis state is controlled by the basis state .

        :param basis_state: [description]
        :type basis_state: int
        :return: [description]
        :rtype: bool
        """
        for i, control_qubit in enumerate(self.control_qubit_indices):
            if (basis_state >> control_qubit) & 1 != self.control_qubit_phases[i]:
                return False
        return True



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-06-22 13:11:39
Last Modified by: Hanyu Wang
Last Modified time: 2023-06-23 00:01:25
"""

from .cx import *
from .mcry import *

from .rx import *
from .ry import *
from .rz import *

from .crx import *
from .cry import *
from .crz import *

from .z import *
from .x import *
from .muxy import *
from .base import *

from .u import *
from .cu import *


from .mcmy import *



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-08-11 23:02:57
Last Modified by: Hanyu Wang
Last Modified time: 2023-08-12 00:02:46
"""

from .basic_gates import *
from .super_gates import *
from .qcircuit import *
from .decomposition import *
